"""
QUESTIONS

This extension offers a new syntaw to write tests and answers.
"""


from ptyx.latex_generator import Compiler


def test_CALC():
    test = r"$#CALC{\dfrac{2}{3}+1}=#RESULT$ et $#CALC[a]{\dfrac{2}{3}-1}=#a$"
    c = Compiler()
    latex = c.parse("#LOAD{geophyx}" + test)
    print(latex)
    assert latex == r"$\dfrac{2}{3}+1=\frac{5}{3}$ et $\dfrac{2}{3}-1=- \frac{1}{3}$"


def test_TABVAR():
    test = (
        "$#{a=2;}\\alpha=#{alpha=3},\\beta=#{beta=5}\n\n"
        "#TABVAR[limites=False,derivee=False]f(x)=#a*(x-#alpha)^2+#beta#END_TABVAR$"
    )
    result = """$\\alpha=3,\\beta=5
\\setlength{\\TVextraheight}{\\baselineskip}
\\[\\begin{tabvar}{|C|CCCCC|}
\\hline
\\,\\,x\\,\\,                            &-\\infty      &        &3&      &+\\infty\\\\
\\hline
\\niveau{1}{2}\\raisebox{0.5em}{$f(x)$}&\\niveau{2}{2}&\\decroit&5&\\croit&\\\\
\\hline
\\end{tabvar}\\]
% x;f(x):(-oo;) >> (3;5) << (+oo;)
% f(x)=2*(x-3)^2+5\n$"""
    c = Compiler()
    latex = c.parse("#LOAD{geophyx}" + test)
    assert latex == result



import random
import functools
from collections import namedtuple
from math import gcd
from typing import Callable, Iterable

from numpy import array

from ptyx.config import param, SYMPY_AVAILABLE


if SYMPY_AVAILABLE:
    S: Callable
    from sympy import S, Integer

    INTEGERS_TYPES: tuple[type, ...] = (int, Integer)
else:

    def S(val):
        return val

    INTEGERS_TYPES = (int,)


# Important note: all the following functions are sandboxed.
# By this, I mean that any external call to random won't affect random state for
# the following functions.

_RANDOM_STATE = random.getstate()
_SANDBOXED_MODE = False


class Point(namedtuple("Point", ["x", "y"])):
    def __add__(self, vec):
        return Point(self.x + vec[0], self.y + vec[1])

    def __neg__(self):
        return Point(-self.x, -self.y)

    def __sub__(self, vec):
        return Point(self.x - vec[0], self.y - vec[1])

    def __rmul__(self, k):
        return Point(k * self.x, k * self.y)


def _print_state():
    """For debuging purpose."""
    print(29 * "*")
    print("randfunc._RANDOM_STATE value:")
    print(hash(_RANDOM_STATE))
    print(29 * "*")


def sandboxed(func):
    @functools.wraps(func)
    def wrapper(*args, **kw):
        global _RANDOM_STATE, _SANDBOXED_MODE
        if not _SANDBOXED_MODE:
            old_state = random.getstate()
            random.setstate(_RANDOM_STATE)
            _SANDBOXED_MODE = True
            try:
                val = func(*args, **kw)
            finally:
                _RANDOM_STATE = random.getstate()
                random.setstate(old_state)
                _SANDBOXED_MODE = False
        else:
            val = func(*args, **kw)
        return val

    return wrapper


def set_seed(value):
    global _RANDOM_STATE
    random.seed(value)
    _RANDOM_STATE = random.getstate()


@sandboxed
def randint(a: int = None, b: int = None, exclude: Iterable = ()) -> int:
    """Generate a random integer between `a` and `b`.

    Forbidden values may be specified using `exclude` argument.
    """
    if a is None and b is not None:
        a, b = b, a
    if b is None:
        b = 9 if a is None else a
        a = 2
    assert a is not None and b is not None
    while a in exclude:
        a += 1
        if a > b:
            raise ValueError("Can't statisfy constraints !")
    while b in exclude:
        b -= 1
        if a > b:
            raise ValueError("Can't statisfy constraints !")
    while True:
        val = random.randint(a, b)
        if val not in exclude:
            break
    if param["sympy_is_default"]:
        val = S(val)
    return val


@sandboxed
def srandint(a: int = None, b: int = None, exclude: Iterable = ()) -> int:
    """Generate a random integer between `a` and `b` or between `-b` and `-a`.

    `a` and `b` should be positive integers.
    Forbidden values may be specified using `exclude` argument.
    """
    if a is None and b is None:
        a, b = 2, 9
    elif a is None:
        a = 2
    elif b is None:
        a, b = 2, a
    assert a is not None and b is not None
    # pylint: disable=invalid-unary-operand-type
    while a in exclude and -a in exclude:
        a += 1
        if a > b:
            raise ValueError("Can't statisfy constraints !")
    while b in exclude and -b in exclude:
        b -= 1
        if a > b:
            raise ValueError("Can't statisfy constraints !")
    while True:
        val = (-1) ** random.randint(0, 1) * randint(a, b)
        if val not in exclude:
            return val


@sandboxed
def randsign():
    val = (-1) ** random.randint(0, 1)
    if param["sympy_is_default"]:
        val = S(val)
    return val


@sandboxed
def randbool():
    return bool(randint(0, 1))


@sandboxed
def randpoint(a=None, b=None, exclude=()):
    while True:
        x = randint(a, b)
        y = randint(a, b)
        if (x, y) not in exclude:
            return Point(x, y)


@sandboxed
def srandpoint(a=None, b=None, exclude=()):
    while True:
        x = srandint(a, b)
        y = srandint(a, b)
        if (x, y) not in exclude:
            return Point(x, y)


def is_mult_2_5(val: int | Iterable[int]) -> bool:
    """Test if integer `val` matches 2**n * 5**m.

    `val` may also be a list (or other iterable) of integers,
     in that case all values must match 2**n * 5**m.
    """
    if hasattr(val, "__iter__"):
        return all(is_mult_2_5(v) for v in val)
    if val == 0 or not isinstance(val, INTEGERS_TYPES):
        return False
    while val % 5 == 0:
        val = val // 5
    while val % 2 == 0:
        val = val // 2
    return val in (1, -1)


@sandboxed
def randfrac(a=None, b=None, exclude=(), not_decimal=False, den=None):
    """Return a random positive fraction which is never an integer.

    By default, numerator and denominators are random integers between `a` and `b`.

    Use `den` to specify denominator value; `den` must be an integer or
    a list of integers.
    """
    if b is None:
        b = 9 if a is None else a
        a = 2
    if hasattr(den, "__iter__"):
        # To allow rando.choice() and multiple iterations.
        den = list(den)
    if (den is None and a in (-1, 0, 1) and b in (-1, 0, 1)) or a == b:
        # This would lead to infinite loop.
        raise ValueError("(%s, %s) are not valid parameters." % (a, b))
    if not_decimal and is_mult_2_5(den):
        raise ValueError("chosen denominator is not compatible with `not_decimal` option.")
    while True:
        if den is None:
            d = randint(a, b)
            if d in (0, 1):
                continue
            n = randint(a, b)
        else:
            n = randint(a, b)
            if hasattr(den, "__iter__"):
                d = random.choice(den)
            else:
                d = den
            if gcd(d, n) not in (-1, 1):
                # XXX this may lead to infinite loop if wrong arguments are passed
                continue
        val = S(n) / S(d)
        if not_decimal and is_mult_2_5(val.q):
            continue
        if not val.is_integer and val not in exclude:
            return val


@sandboxed
def srandfrac(a=None, b=None, exclude=(), not_decimal=False, den=None):
    """Return a random signed fraction which is never an integer.

    Use `den` to specify denominator value; `den` must be an integer or a tuple
    of integers.
    """
    while True:
        val = (-1) ** randint(0, 1) * randfrac(a, b, not_decimal=not_decimal, den=den)
        if val not in exclude:
            return val


@sandboxed
def randchoice(items, *others, **kw):
    """Select randomly an item.

    Note that `randchoice(1, 2, 3)` is equivalent to `randchoice([1, 2, 3])`.
    """
    if others:
        # `items` is then only the first element.
        items = [items] + list(others)
    if kw.get("signed"):
        items = list(items) + [-1 * item for item in items]
    if "exclude" in kw:
        items = [val for val in items if val not in kw["exclude"]]
    val = random.choice(items)
    if isinstance(val, (int, float, complex)):
        val = S(val)
    return val


@sandboxed
def randpop(list_or_set):
    """Randomely remove an item from list or set and return it."""
    i = random.randint(0, len(list_or_set) - 1)
    if isinstance(list_or_set, list):
        return list_or_set.pop(i)
    elif isinstance(list_or_set, set):
        for j, v in enumerate(list_or_set):
            if i == j:
                list_or_set.remove(v)
                return v
    else:
        raise NotImplementedError


@sandboxed
def srandchoice(*items, **kw):
    kw["signed"] = True
    return randchoice(*items, **kw)


@sandboxed
def shuffle(items: list, _random: Callable[[], float] = None) -> None:
    random.shuffle(items, _random)


@sandboxed
def randmaketrans(string, _random=None):
    """Shuffle string letters and return a random translation table usable for str.translate()."""
    letters = list(string)
    if len(letters) != len(set(letters)):
        raise ValueError(f"Same letter appears twice in {string!r}.")
    random.shuffle(letters, _random)
    shuffled_string = "".join(letters)
    return str.maketrans(string, shuffled_string)


@sandboxed
def randfloat(a, b, d=5, exclude=()):
    k = 10**d
    exclude = {int(round(v * k)) for v in exclude}
    while True:
        n = randint(int(round(a * k)) + 1, int(round(b * k)) - 1)
        if n not in exclude:
            return float(n / k)


@sandboxed
def srandfloat(a, b, d=5, exclude=()):
    return float(randsign()) * randfloat(a, b, d, exclude)


@sandboxed
def randmatrix(size=(3, 3), rank=None, unique=False, func=srandint, **kw):
    """Return a matrix of dimensions `size` (number of lines, number of columns).

    `func` is used to build the random values, and **kw are passed to it.
    Be careful with `rank`, as it may induce infinite recursion if
    condition can't be satisfied.

    Set `unique` to True if each coefficient must be unique.
    Note that you can't specify rank then.
    """
    from sympy import Matrix

    if rank is None:
        m, n = size
        return Matrix(m, n, many(m * n, func=func, unique=unique, **kw))
    elif unique:
        raise NotImplementedError("You can't specify rank if you set unique=True.")
    elif rank > min(size):
        raise ValueError("Matrix rank can't exceed lines nor columns number.")
    else:
        while True:
            # noinspection PyUnusedLocal
            matrix = [[func(**kw) for j in range(size[1])] for i in range(rank)]
            if Matrix(matrix).rank() == rank:
                break
        while len(matrix) < size[0]:
            matrix.append(list(sum(srandint() * array(line) for line in matrix)))
    return Matrix(matrix)


def many(n=2, func=srandint, unique=True, **kw):
    """Return several numbers at once.

    By default, every number is unique.
    Note this can lead to infinite recursion if n is too large and
    `unique` is set to True (default value)."""
    values = []
    kw.setdefault("exclude", [])
    # Make a copy of `exclude`, to not modify the given list.
    kw["exclude"] = list(kw["exclude"])
    for i in range(n):
        val = func(**kw)
        if unique:
            kw["exclude"].append(val)
        values.append(val)
    return values


def distinct(*vals):
    """Test that all values are distinct.

    Values must be either immutable or convertible to tuples::

    >>> distinct(1, 2, 3)
    True
    >>> distinct(1, 1, 3)
    False
    >>> distinct(5, 4, 5.0)
    False
    >>> distinct((1, 2), {1, 3})
    True
    >>> distinct((1, 2), {1, 2})
    False
    """
    try:
        return len(set(vals)) == len(vals)
    except TypeError:
        hashable_vals = set()
        for val in vals:
            try:
                hashable_vals.add(val)
            except TypeError:
                try:
                    hashable_vals.add(tuple(val))
                except TypeError:
                    raise ValueError(f"Unsupported type {type(val)} for {val!r}")
        return len(set(hashable_vals)) == len(vals)



import random
import re
import traceback
from importlib import import_module, metadata
from os.path import dirname, basename, join
from pathlib import Path
from types import ModuleType
from typing import Optional, Union, Callable, Iterable, Dict, Tuple, List, TypedDict

from ptyx.extensions import CompilerExtension

import ptyx.randfunc as randfunc
from ptyx import __version__
from ptyx.config import param, SYMPY_AVAILABLE
from ptyx.context import GLOBAL_CONTEXT

# from ptyx.printers import sympy2latex
from ptyx.syntax_tree import Node, SyntaxTreeGenerator, Tag, TagSyntax
from ptyx.utilities import advanced_split, numbers_to_floats, _float_me_if_you_can, latex_verbatim


class State(TypedDict, total=False):
    syntax_tree: Node
    seed: Optional[int]
    input: str
    path: Optional[Path]
    loaded_extensions: dict[str, ModuleType]
    plain_ptyx_code: str
    after_include: Optional[str]


# =============================================================================
# #ASSERT{}
# #CASE{int}...#CASE{int}...#END
# #COMMENT...#END
# #DEBUG
# #IF{bool}...#ELIF{bool}...#ELSE...#END
# #IFNUM{int}{...}
# #MACRO{name}
# #NEW_MACRO{name}...#END
# #PICK[option,int]{list}
# #PYTHON...#END
# #RAND[option,int]{list}
# #SEED{int}
# #SHUFFLE #ITEM...#ITEM...#END
# #SIGN
# #SYMPY...#END
# #TEST{bool}{...}{...}
# #+, #-,  #*, #=, #? (special operators)
# #varname or #[option,int]varname
# ## options : sympy, python, float
# ## int : round
#
# =============================================================================


# noinspection PyPep8Naming,PyMethodMayBeStatic
class LatexGenerator:
    """Convert text containing ptyx tags to plain LaTeX."""

    convert_tags = {
        "+": "ADD",
        "-": "SUB",
        "*": "MUL",
        "=": "EQUAL",
        "?": "SIGN",
        "#": "SHARP",
    }

    # noinspection RegExpRedundantEscape
    re_varname = re.compile(r"[A-Za-z_]\w*(\[.+\])?$")

    def __init__(self, compiler=None):
        self.clear()
        # Some pTyX code may be generated dynamically inside a #PYTHON code
        # block for example. So, LatexGenerator need his own STG (Syntax Tree Generator)
        # to be able to scan code at runtime.
        # (Every run implies generating a syntax tree again for each
        # of these pieces of embedded pTyX code.)
        self.parser = SyntaxTreeGenerator()
        # Access to compiler is needed by some extensions.
        self.compiler = compiler
        # Add ability to share values between compilations (needed by some
        # extensions).
        self.cache = {}

    def clear(self):
        self.macros = {}
        self.context = GLOBAL_CONTEXT.copy()
        self.context["PTYX_LATEX"] = []
        self.backups = []
        # When write() is called from inside a #PYTHON ... #END code block,
        # its argument may contain pTyX code needing parsing.
        # Write argument used to be parsed by default, but this was rarely needed
        # and led to very surprising behaviour sometimes from user perspective.
        # Now, one should use write(..., parse=True) if needed, which is much saner.
        self.context["write"] = self.write
        # Internal flags
        self.flags = {}

    def reset(self):
        """To overwrite."""
        self.clear()

    def set_new_context(self, context: Optional[Dict] = None):
        """Set a new context of evaluation for code, except for PTYX_LATEX value."""
        if context is None:
            context = GLOBAL_CONTEXT
        # Use deepcopy instead ?
        context = context.copy()
        # Copy internal parameters to new context.
        for key in self.context:
            if key.startswith("PTYX_"):
                context[key] = self.context[key]
        self.context = context

    @property
    def NUM(self):
        return self.context["PTYX_NUM"]

    @property
    def WITH_ANSWERS(self):
        return self.context.get("PTYX_WITH_ANSWERS")

    def _open_temp_context(self):
        # Store generated text in a temporary location, instead of self.context['LATEX'].
        # Function `function` will then be applied to this text,
        # before text is appended to self.context['LATEX'].
        # Backups may need to be read when parsing #= special tag,
        # so store them in `self.backups`.
        self.backups.append(self.context["PTYX_LATEX"])
        self.context["PTYX_LATEX"] = []

    def _apply_func_and_close_temp_context(self, function, **options):
        code = "".join(self.context["PTYX_LATEX"])
        if callable(function):
            function = [function]
        for f in function:
            code = f(code, **options)
        self.context["PTYX_LATEX"] = self.backups.pop()
        self.write(code)

    def parse_node(
        self,
        node: Node,
        function: Optional[Union[Callable, Iterable[Callable]]] = None,
        **options,
    ):
        """Parse a node in a pTyX syntax tree.

        Return True if block content was recursively parsed, and False else.

        In particular, when parsing an IF block, True will be returned if and
        only if the block condition was satisfied.
        """
        if function is not None:
            self._open_temp_context()
        name = node.name
        assert isinstance(name, str), repr(node)
        tag = self.convert_tags.get(name, name)
        try:
            method = getattr(self, f"_parse_{tag}_tag")
            return method(node)
        except AttributeError:
            print(f"Error: method '_parse_{tag}_tag' not found.")
            raise
        except Exception:
            print(f"Error when calling method '_parse_{tag}_tag'.")
            raise
        finally:
            if function is not None:
                self._apply_func_and_close_temp_context(function, **options)

    def _parse_children(
        self,
        children: Iterable[Union[str, Node]],
        function: Optional[Union[Callable, Iterable[Callable]]] = None,
        **options,
    ):
        """Parse all children nodes.

        Resulting LaTeX code will be appended to `self.context['PTYX_LATEX']`.

        If `function` is not None, apply `function` to resulting LaTeX
        code before appending it to `self.context['PTYX_LATEX']`.
        (So, `function` signature must be: function(str, **options) -> str).

        Note that `function` can also be a list of functions.
        In that case, the first function of the list is also the first
        to be applied.
        """

        if function is not None:
            self._open_temp_context()

        for child in children:
            if isinstance(child, str):
                self.write(child)
            else:
                # All remaining children should be nodes now.
                assert isinstance(child, Node)
                # Nodes are either numbered, or have a name.
                # Numbered nodes correspond to command arguments. Those should
                # have been processed before, and not be passed to _parse_children().
                assert isinstance(child.name, str), (
                    f"Argument {child.name!r} should have been processed "
                    "and removed before calling _parse_children() !"
                )

                self.parse_node(child)

        if function is not None:
            self._apply_func_and_close_temp_context(function, **options)

    @staticmethod
    def _parse_options(node: Node):
        """Parse a tag options, following the syntax {key1=val1,...}."""
        options = node.options
        args = []
        kw = {}
        if options is not None:
            options_list = options.split(",")
            for option in options_list:
                if "=" in option:
                    key, val = option.split("=")
                    kw[key.strip()] = val.strip()
                else:
                    args.append(option.strip())
        return args, kw

    def write(self, text: str, parse: bool = False, verbatim: bool = False):
        """Append a piece of LaTeX text to context['PTYX_LATEX'].

        :param text: a block of text, which may contain pTyX code.
        :type text: string
        :param parse: indicate text have to be parsed.
        :type parse: bool
        :param verbatim: emulate LaTeX verbatim mode (escape LaTeX characters, keep indentation...)
        :type verbatim: bool

        .. note:: Param `parse` defaults to False, since in most cases text is already
                  parsed at this state.
                  As an exception, when called from inside a #PYTHON [...] #END
                  code block, write() may be applied to unparsed text.
        """
        # Parsing at this stage is quite expansive yet unnecessary most of the time,
        # so some basic testing is done before.
        text = str(text)
        if text.strip():
            for flag in "+-*":
                if self.flags.get(flag):
                    text = (r"\times " if flag == "*" else flag) + text
                    self.flags[flag] = False
        if parse and "#" in text:
            if param["debug"]:
                print("Parsing %s..." % repr(text))
            self.parse_node(self.parser.generate_tree(text), function=(latex_verbatim if verbatim else None))
        else:
            if verbatim:
                text = latex_verbatim(text)
            self.context["PTYX_LATEX"].append(text)

    def read(self):
        return "".join(self.context["PTYX_LATEX"])

    def _parse_APART_tag(self, node: Node):
        """Interpret a piece of code in a sandbox, eliminating side effects."""
        # Backup local variables and reset context.
        context_backup = self.context
        self.set_new_context()
        # Interprete code
        self._parse_children(node.children)
        # Restore local variables and update generated LaTeX code.
        self.set_new_context(context_backup)

    def _parse_PTYX_VERSION_tag(self, node: Node):
        def version_tuple(version_: str):
            return version_.split(".")

        assert isinstance(node.children[0], Node), repr(node)
        assert isinstance(node.children[0].children[0], str), repr(node)
        version = version_tuple(node.children[0].children[0])
        if version_tuple(__version__) < version:
            print("Warning: pTyX engine is too old (v%s required)." % version)
        self.context["PTYX_VERSION"] = version

    def _parse_ASK_tag(self, node: Node):
        self._parse_children(node.children, function=self.context.get("format_ask"))

    def _parse_ASK_ONLY_tag(self, node: Node):
        if not self.WITH_ANSWERS:
            self._parse_children(node.children, function=self.context.get("format_ask_only"))
        else:
            print("Skipping ASK_ONLY section...")

    def _parse_ANS_tag(self, node: Node):
        if self.WITH_ANSWERS:
            self._parse_children(node.children, function=self.context.get("format_ans"))

    def _parse_ANSWER_tag(self, node: Node):
        if self.WITH_ANSWERS:
            assert isinstance(node.children[0], Node), repr(node)
            self._parse_children(node.children[0].children, function=self.context.get("format_answer"))

    def _parse_QUESTION_tag(self, node: Node):
        if not self.WITH_ANSWERS:
            assert isinstance(node.children[0], Node), repr(node)
            self._parse_children(node.children[0].children)

    def _parse_IF_tag(self, node: Node):
        test = eval(node.arg(0), self.context)
        if test:
            self._parse_children(node.children[1:])
        return test

    _parse_ELIF_tag = _parse_IF_tag

    def _parse_CONDITIONAL_BLOCK_tag(self, node: Node):
        for child in node.children:
            assert isinstance(child, Node)
            if self.parse_node(child):
                # If an `IF` or `ELIF` node was processed, all successive `ELIF`
                # or `ELSE` nodes must be skipped.
                # (The same for `CASE`).
                break

    def _parse_ELSE_tag(self, node: Node):
        self._parse_children(node.children)

    def _parse_IFNUM_tag(self, node: Node):
        if eval(node.arg(0), self.context) == self.NUM:
            assert isinstance(node.children[1], Node), repr(node)
            self._parse_children(node.children[1].children)

    def _parse_CASE_tag(self, node: Node):
        test = eval(node.arg(0), self.context) == self.NUM
        if test:
            self._parse_children(node.children[1:])
        return test

    def _parse_IMPORT_tag(self, node: Node):
        exec("from %s import *" % node.arg(0), self.context)

    def _parse_LOAD_tag(self, node: Node):
        # LOAD tag is used to load extensions **before** syntax tree is built.
        pass

    def _parse_INCLUDE_tag(self, node: Node):
        # INCLUDE tag is used to insert code in isolated mode **before** syntax tree is built.
        pass

    @staticmethod
    def _format_python_code_snippet(python_code) -> List[str]:
        """Return a list of prettified lines of python code, ready to be printed."""
        msg = ["", "%s %s Executing following python code:" % (chr(9474), chr(9998))]
        lines = [""] + python_code.split("\n") + [""]
        zfill = len(str(len(lines)))
        msg.extend(
            "%s %s %s %s" % (chr(9474), str(i).zfill(zfill), chr(9474), line) for i, line in enumerate(lines)
        )
        n = max(len(s) for s in msg)
        msg.insert(1, chr(9581) + n * chr(9472))
        msg.insert(3, chr(9500) + n * chr(9472))
        msg.append(chr(9584) + n * chr(9472))
        assert isinstance(python_code, str)
        return msg

    def _parse_PYTHON_tag(self, node: Node):
        assert len(node.children) == 1
        python_code = node.children[0]
        assert isinstance(python_code, str), repr(python_code)
        self._exec_python_code(python_code, self.context)

    # Remove comments before generating tree ?
    def _parse_COMMENT_tag(self, node: Node):
        pass

    def _parse_ASSERT_tag(self, node: Node):
        code = node.arg(0)
        test = eval(code, self.context)
        if not test:
            print("Error in assertion (NUM=%s):" % self.NUM)
            print("***")
            print(code)
            print("***")
            assert test

    def _parse_EVAL_tag(self, node: Node):
        args, kw = self._parse_options(node)
        if self.context.get("ALL_FLOATS"):
            self.flags["floats"] = True
        for arg in args:
            if arg.isdigit():
                self.flags["round"] = int(arg)
            elif arg == ".":
                self.flags["."] = True
            elif arg in ("floats", "float"):
                self.flags["floats"] = True

            elif arg == "str":
                self.flags["str"] = True
            else:
                raise ValueError("Unknown flag: " + repr(arg))
        # XXX: support options round, float, (sympy, python,) select and rand
        code = node.arg(0)
        assert isinstance(code, str), type(code)
        try:
            txt = self._eval_and_format_python_expr(code)
        except Exception:
            print("ERROR: Can't evaluate this: " + repr(code))
            raise
        # Tags must be cleared *before* calling .write(txt), since .write(txt)
        # add '+', '-' and '\times ' before txt if corresponding flags are set,
        # and ._eval_and_format_python_expr() has already do this.
        self.flags.clear()
        self.write(txt)

    def _parse_MACRO_tag(self, node: Node):
        name = node.arg(0)
        self.macros[name] = node.children[1:]

    def _parse_CALL_tag(self, node: Node):
        """Calling a macro."""
        name = node.arg(0)
        if name not in self.macros:
            raise NameError(f"Error: MACRO {name!r} undefined.")
        self._parse_children(self.macros[name])

    def _parse_ENUM_tag(self, node: Node):
        # This tag does nothing by itself, but is used by some extensions.
        self._parse_children(node.children)

    def _parse_VERBATIM_tag(self, node: Node):
        self._parse_children(node.children, function=latex_verbatim)

    @staticmethod
    def _child_index(children, name):
        """Return the index of the first child of name `name` in `children` list.

        Argument `name` must be a `Node` name.

        A `ValueError` is raised if no such `Node` is found.
        """
        for i, child in enumerate(children):
            if isinstance(child, Node) and child.name == name:
                return i
        raise ValueError(f"No {name} Node found.")

    def _shuffle_and_parse_children(self, node: Node, children=None, target: str = "ITEM", **kw):
        # Shuffles all the #ITEM sections inside a #SHUFFLE block.
        # Note that they may be some text or nodes before first #ITEM,
        # if so they should be left unmodified at first position.
        # Nota: children may be only some children of given node.
        # This is used when some children were already processed before
        # (this is the case for arguments usually).
        if children is None:
            children = node.children
        if not children:
            return
        try:
            i = self._child_index(children, target)
        except ValueError:
            # `target` not found: nothing to shuffle.
            print(f"WARNING: Nothing to shuffle ({target!r} not found).")
            self._parse_children(children, **kw)
            return
        items = children[i:]
        # Now, we will shuffle only items which are nodes of name `target`.
        # So, we split items in groups, so that every group starts with
        # a Node named `target`.
        # For example, if target is `QUESTION`, and items are
        # ['QUESTION', 'OTHER', 'QUESTION', 'OTHER', 'ANOTHER'],
        # groups will be [['QUESTION', 'OTHER'],
        #                 ['QUESTION', 'OTHER', 'ANOTHER']]
        # Then, we will shuffle those groups. We may obtain for example :
        #                [['QUESTION', 'OTHER', 'ANOTHER'],
        #                 ['QUESTION', 'OTHER']]
        # Finally, we flatten `groups` list to obtain the following:
        # ['QUESTION', 'OTHER', 'ANOTHER', 'QUESTION', 'OTHER']
        # 1. Generate groups
        groups = []
        for item in items:
            if isinstance(item, Node) and item.name == target:
                groups.append([item])
            else:
                groups[-1].append(item)
        # 2. Shuffle groups
        randfunc.shuffle(groups)
        #        print(groups)
        #        input('-- pause --')
        # 3. Flatten `groups` list
        items = sum(groups, [])

        #        print(f'items: {items}')
        #        input('-- pause --')
        self._parse_children(children[:i] + items, **kw)
        # print('\n------------')
        # print('SHUFFLE: %s elements, excluding first %s.' % len(children), i)
        # print('state hash is %s' % hash(random.getstate()))
        # print('------------\n')

    def _parse_SHUFFLE_tag(self, node: Node):
        self._shuffle_and_parse_children(node)

    def _parse_ITEM_tag(self, node: Node):
        self._parse_children(node.children)

    def _parse_SEED_tag(self, node: Node):
        # SEED tag is a special tag which is managed independently, at first pass
        # (it avoids user including it inadvertently inside
        # a conditional block or a #ASK_ONLY/#END block, which may result in
        # what seems to be a very strange behaviour of the compiler !)
        pass

    def _pick_and_parse_children(self, node: Node, children: List = None, target: str = "ITEM", **kw):
        # Choose only one between all the #ITEM sections inside a #PICK block.
        # Note that they may be some text or nodes before first #ITEM,
        # if so they should be left unmodified at their original position.
        if children is None:
            children = node.children
        if not children:
            return
        i = self._child_index(children, target)
        items = children[i:]
        for item in items:
            if not (isinstance(item, Node) and item.name == target):
                log = [
                    "This is current structure:",
                    node.display(),
                    rf"\n{item!r} is not an {target!r} node !",
                ]
                raise RuntimeError("\n".join(log))
        item = randfunc.randchoice(items)
        self._parse_children(children[:i] + [item], **kw)
        # print('\n------------')
        # print('SHUFFLE: %s elements, excluding first %s.' % len(children), i)
        # print('state hash is %s' % hash(random.getstate()))
        # print('------------\n')

    def _parse_PICK_tag(self, node: Node) -> None:
        self._pick_and_parse_children(node)

    # TODO: Refactor _parse_PICK_tag/_parse_SHUFFLE_tag

    def _parse_ROOT_tag(self, node: Node) -> None:
        self._parse_children(node.children)

    def _parse_FREEZE_RANDOM_STATE_tag(self, node: Node) -> None:
        state = random.getstate()
        self._parse_children(node.children)
        random.setstate(state)

    def _parse_TEST_tag(self, node: Node) -> None:
        try:
            if eval(node.arg(0), self.context):
                assert isinstance(node.children[1], Node), repr(node)
                self._parse_children(node.children[1].children)
            else:
                assert isinstance(node.children[2], Node), repr(node)
                self._parse_children(node.children[2].children)
        except Exception:
            print(node.display(color=False))
            raise

    def _parse_SHARP_tag(self, node: Node) -> None:
        """2 sharps (##) will be converted to 1 sharp (#)"""
        self.write("#")

    def _parse_ADD_tag(self, node: Node) -> None:
        # a '+' will be displayed at the beginning of the next result if positive ;
        # if the result is negative, nothing will be done, and if null,
        # no result at all will be displayed.
        self.flags["+"] = True

    def _parse_SUB_tag(self, node: Node) -> None:
        # a '-' will be displayed at the beginning of the next result, and the result
        # will be embedded in parentheses if negative.
        self.flags["-"] = True

    def _parse_MUL_tag(self, node: Node) -> None:
        # a '\times' will be displayed at the beginning of the next result, and the result
        # will be embedded in parentheses if negative.
        self.flags["*"] = True

    def _parse_EQUAL_tag(self, node: Node) -> None:
        # Display '=' or '\approx' when a rounded result is requested :
        # if rounded is equal to exact one, '=' is displayed.
        # Else, '\approx' is displayed instead.
        self.flags["="] = True
        # All other operations (#+, #-, #*) occur just before number, but between `=` and
        # the result, some formatting instructions may occur (like '\fbox{' for example).
        # So, `#=` is used as a temporary marker, and will be replaced by '=' or '\approx' later.
        self.write("#=")

    def _parse_SIGN_tag(self, node: Node) -> None:
        # '>0' or '<0' will be displayed after the next result, depending on it's sign.
        # (If result is zero, this won't do anything.)
        last_value = self.context["_"]
        if last_value > 0:
            self.write(">0")
        elif last_value < 0:
            self.write("<0")

    def _parse_DEBUG_tag(self, node: Optional[Node]) -> None:
        while True:
            msg = "Debug point. Enter command, or quit (q! + ENTER):"
            sep = len(msg) * "="
            print(sep)
            print(msg)
            print(sep)
            command = input("[In]: ")
            print("[Out]: ")
            if command == "q!":
                break
            else:
                try:
                    exec(command, self.context)
                except Exception as e:
                    print("*** ERROR ***")
                    print(e)

    def _parse_PRINT_tag(self, node: Node) -> None:
        print(node.arg(0))

    @staticmethod
    def _exec(code, context):
        """exec is encapsulated in this function so as to avoid problems
        with free variables in nested functions."""
        exec(code, context)

    def _exec_python_code(self, code: str, context: dict):
        code = code.replace("\r", "")
        code = code.rstrip().lstrip("\n")
        msg = self._format_python_code_snippet(code)
        # Indentation test
        initial_indent = len(code) - len(code.lstrip(" "))
        if initial_indent:
            # remove initial indentation
            code = "\n".join(line[initial_indent:] for line in code.split("\n"))
        try:
            self._exec(code, context)
        except Exception as e:  # noqa
            for tb in traceback.extract_tb(e.__traceback__):
                if tb.name == "<module>" and isinstance(tb.lineno, int):
                    i = tb.lineno + 4
                    msg[i] = f"\u001b[33m{msg[i]}\u001b[0m"
                    break
            e.msg = "\n".join(msg)  # type: ignore
            raise

        return code

    def _eval_python_expr(self, code: str):
        flags = self.flags
        context = self.context
        if not code:
            return ""
        sympy_code = flags.get("sympy", param["sympy_is_default"])

        if sympy_code and not SYMPY_AVAILABLE:
            raise ImportError("sympy library not found.")

        varname = ""
        i = code.find("=")
        if 0 < i < len(code) - 1 and code[i + 1] != "=":
            varname = code[:i].strip()
            if re.match(self.re_varname, varname):
                # This is (probably) a variable name (or a list or dict item reference).
                code = code[i + 1 :]
            else:
                # This is not a variable name.
                varname = ""
        # Last value will be accessible through '_' variable
        if not varname:
            varname = "_"
        if " if " in code and " else " not in code:
            code += " else ''"
        if SYMPY_AVAILABLE and sympy_code:
            import sympy

            try:
                result = sympy.sympify(code, locals=context)
                if isinstance(result, str):
                    result = result.replace("**", "^")
            except (sympy.SympifyError, AttributeError):
                # sympy.sympify() can't parse attributes and methods inside
                # code for now (AttributeError is raised then).
                result = eval(code, context)
                print("Warning: sympy can't parse %s. " "Switching to standard evaluation mode." % repr(code))
            except Exception:
                # print sorted(context.keys())
                print("Uncatched error when evaluating %s" % repr(code))
                raise
        else:
            result = eval(code, context)
        result = context["_"] = self._apply_flag(result)
        i = varname.find("[")
        # for example, varname == 'mylist[i]' or 'mylist[2]'
        context["LAST"] = result
        if i == -1:
            context[varname] = result
        else:
            key = eval(varname[i + 1 : -1], context)
            varname = varname[:i].strip()
            context[varname][key] = result
        return result

    def _eval_and_format_python_expr(self, code: str) -> str:
        flags = self.flags
        context = self.context
        if not code:
            return ""
        sympy_code = flags.get("sympy", param["sympy_is_default"])

        display_result = True
        if code.endswith(";"):
            code = code.rstrip(";")
            display_result = False

        for subcode in advanced_split(code, ";", brackets=()):
            result = self._eval_python_expr(subcode)
        # Note that only last result will be displayed.
        # In particular, if code ends with ';', last result will be ''.
        # So, '#{a=5}' and '#{a=5;}' will both affect 5 to `a`,
        # but the second will not display '5' on final document.

        if not display_result:
            return ""

        if sympy_code and not flags.get("str"):
            from ptyx.printers import sympy2latex

            latex = sympy2latex(result, **flags)
        else:
            latex = str(result)

        def neg(latex):
            return latex.lstrip().startswith("-")

        if flags.get("+"):
            if result == 0:
                latex = ""
            elif not neg(latex):
                latex = "+" + latex
        elif flags.get("*"):
            if neg(latex) or getattr(result, "is_Add", False):
                latex = r"\left(" + latex + r"\right)"
            latex = r"\times " + latex
        elif flags.get("-"):
            if neg(latex) or getattr(result, "is_Add", False):
                latex = r"\left(" + latex + r"\right)"
            latex = "-" + latex
        elif flags.get("="):
            if flags.get("result_is_exact"):
                symb = " = "
            else:
                symb = r" \approx "
            # Search backward for temporary `#=` marker in list, and replace
            # by appropriate symbol.
            for textlist in self.backups + [context["PTYX_LATEX"]]:
                for i, elt in enumerate(reversed(textlist)):
                    if elt == "#=":
                        textlist[len(textlist) - i - 1] = symb
                        return latex
            print("Debug warning: `#=` couldn't be found when scanning context !")
            print("There is most probably a bug in pTyX, entering debugging...")
            self._parse_DEBUG_tag(None)
        return latex

    def _apply_flag(self, result):
        """Apply special parameters like [num], [rand], [floats] or [round] to result.

        Note that [num] and [rand] parameters require that result is iterable,
        otherwise, nothing occurs.
        If result is iterable, an element of result is returned, chosen according
        to current flag."""
        flags = self.flags
        if hasattr(result, "__iter__"):
            if flags.get("rand"):
                result = random.choice(result)
            elif flags.get("select"):
                result = result[self.NUM % len(result)]

        if "round" in flags:
            try:
                if SYMPY_AVAILABLE:
                    round_result = numbers_to_floats(result, ndigits=flags["round"])
                else:
                    round_result = round(result, flags["round"])
            except ValueError:
                print("** ERROR while rounding value: **")
                print(result)
                print("-----")
                print("".join(self.context["PTYX_LATEX"])[-100:])
                print("-----")
                raise
            if SYMPY_AVAILABLE:
                import sympy
            # noinspection PyUnboundLocalVariable
            if SYMPY_AVAILABLE and isinstance(result, sympy.Basic):
                flags["result_is_exact"] = {_float_me_if_you_can(elt) for elt in result.atoms()} == {
                    _float_me_if_you_can(elt) for elt in round_result.atoms()
                }
            else:
                flags["result_is_exact"] = result == round_result
            result = round_result
        elif "floats" in self.flags:
            result = numbers_to_floats(result)
        return result


class Compiler:
    """Compiler is the main object of pTyX.

    Usage:

        >>> from ptyx.latex_generator import Compiler
        >>> c = Compiler()
        >>> c.parse('#{a=S(1)/2} + #{b=2} = #{a+b}')
        '\frac{1}{2} + 2 = \frac{5}{2}'

    The following methods are called successively to generate LaTeX code from
    a pTyX file:
        * .read_file(path) will read the content of the file.
        * ._include_subfiles() will parse #INCLUDE{filename} tags and
          include coresponding files content.
        * ._call_extensions() will search for extensions in this content,
          then call the extensions to convert content into plain pTyX code.
        * ._read_seed() will search for a #SEED tag, or give a default seed
          value, used to generate all pseudo-random content later.
        * .generate_syntax_tree() will convert this code into a syntax tree.
          This should be done only once ofr each document, even if multiple
          versions of this document are needed.
        * Finally, .get_latex() will generate and return the LaTeX code.

      Pseudo-random content will depend on the seed (see above), but also
      of the document number, given by `gen.context['PTYX_NUM']`.
      So, changing `gen.context['PTYX_NUM']` enables to generate different
      versions of the same document.
    """

    def __init__(self):
        self.syntax_tree_generator = SyntaxTreeGenerator()
        self.latex_generator = LatexGenerator(self)
        self.reset()

    def reset(self) -> None:
        self._state: State = {}
        # Make SyntaxTreeGenerator context free ?
        self.syntax_tree_generator.reset()
        self.latex_generator.reset()

    def read_code(self, code: str) -> None:
        """Feed compiler with given code."""
        self._state["path"] = None
        self._state["input"] = code

    def read_file(self, path: Union[Path, str]) -> None:
        """Feed compiler with given file code."""
        self._state["path"] = Path(path).expanduser().resolve()
        with open(path, "r") as input_file:
            self._state["input"] = input_file.read()

    @property
    def dir_path(self) -> Path:
        """Return input ptyx file directory, if any, or current working directory else."""
        file_path = self.file_path
        return Path.cwd() if file_path is None else file_path.parent

    def _resolve_path(self, path: Union[str, Path]) -> Path:
        """Interpret `path` relatively to input ptyx file."""
        if isinstance(path, str):
            path = Path(path.strip())
        path = path.expanduser()  # do NOT resolve yet !
        if not path.is_absolute():
            path = self.dir_path / path
        return path

    def _include_subfiles(self, code: str) -> str:
        """Parse all #INCLUDE tags, then include corresponding files content."""

        def include(match):
            path = self._resolve_path(match.group(1))
            with open(path) as file:
                return f"\n#APART\n{file.read()}#END_APART\n"

        # noinspection RegExpRedundantEscape
        return re.sub(r"#INCLUDE\{([^}]+)\}", include, code)

    def _call_extensions(self, code: str) -> tuple[str, dict[str, ModuleType]]:
        """Search for extensions (#LOAD{name} tags), then call them."""
        # First, we search if some extensions must be load.
        # This must be done at the very beginning, since extensions may
        # define their own specialized language, to be converted to
        # valid pTyX code (and then to LaTeX).
        names = []

        def collect(match):
            names.append(match.group(1))
            return ""

        # Use re.sub to find all extensions and remove #LOAD tags in one pass.
        # noinspection RegExpRedundantEscape
        code = re.sub(r"#LOAD\{\s*(\w+)\s*\}", collect, code)
        extensions: Dict[Tag, ModuleType] = {}
        tags_syntax: Dict[Tag, TagSyntax] = {}
        tags_source: Dict[Tag, str] = {}
        latex_generator_extensions = []
        for extension_name in names:
            try:
                extensions[extension_name] = import_module(f"ptyx.extensions.{extension_name}")
            except ImportError:
                # Try to find a matching registered plugin.
                for entry_point in metadata.entry_points(group="ptyx.extensions"):
                    if entry_point.name == extension_name:
                        extensions[extension_name] = import_module(entry_point.value)
                        break
                else:
                    traceback.print_exc()
                    raise ImportError(f"Extension {extension_name} not found.")
            try:
                # This extension may define a function `extend_compiler()` to customize the compiler.
                extensions_dict: CompilerExtension = getattr(extensions[extension_name], "extend_compiler")()
                # This extension may subclass LatexGenerator class, notably to handle new tags.
                if "latex_generator" in extensions_dict:
                    latex_generator_extensions.append(extensions_dict["latex_generator"])
                # Test if this extension defines new tags.
                ext_tags = extensions_dict.get("tags", {})
                for tag, syntax in ext_tags.items():
                    # Test for conflict between extensions.
                    if tag in tags_source:
                        raise NameError(
                            f"Extension {extension_name} tries to define tag {tag}, "
                            f"which was already defined by extension {tags_source[tag]}."
                        )
                    # Tag not already declared, everything seems OK.
                    tags_source[tag] = extension_name
                    tags_syntax[tag] = syntax
            except AttributeError:
                pass
        # Load new tags.
        self.add_new_tags(*tags_syntax.items())
        # Update LatexGenerator.
        if latex_generator_extensions:
            # TODO: Test for conflicting methods ?
            class CustomLatexGenerator(*reversed(latex_generator_extensions)):  # type: ignore
                pass

            self.latex_generator.__class__ = CustomLatexGenerator
        for name in names:
            # execute `main()` function of extension.
            if hasattr(extensions[name], "main"):
                code = extensions[name].main(code, self)
        return code, extensions

    def _read_seed(self, code: str) -> Tuple[str, Optional[int]]:
        """Extract seed value from code, searching for #SEED{num} tag.

        Return the code without the #SEED{...} tag, and the seed value (if any, `None` else).
        """
        counter = 0
        value = None

        def seed(match):
            nonlocal counter, value
            value = int(match.group(1))
            counter += 1
            return ""

        # noinspection RegExpRedundantEscape
        code = re.sub(r"#SEED\{\s*(\d+)\s*\}", seed, code)
        if counter == 0:
            path = self._state.get("path")
            print(f"Warning: #SEED not found, using hash of ptyx file path ({path!r}) as seed.")
            value = hash(path)
        elif counter > 1:
            print(f"Warning: multiple #SEED found, only last one will be used: {value}.")
        return code, value

    @staticmethod
    def _assert_no_comment(code: str) -> None:
        """AssertionError raised if comments remain."""
        comments = any((" # " in code, "\n# " in code, code.startswith("# ")))
        if comments:
            print(code)
        assert not comments, "There should be no remaining comment. Maybe a problem with an extension ?"

    def preparse(self) -> None:
        code = self._state.get("input")
        if code is None:
            raise RuntimeError("Compiler.read_code() or Compiler.read_file() must be run first.")
        remove_comments = self.syntax_tree_generator.remove_comments
        # Remove comments first, so one can comment a file inclusion for example.
        code = remove_comments(code)
        assert isinstance(code, str)
        self._assert_no_comment(code)
        code = remove_comments(self._include_subfiles(code))
        self._state["after_include"] = code
        assert isinstance(code, str)
        code, extensions = self._call_extensions(code)
        code, seed = self._read_seed(code)
        # Remove any comment that may have been generated by an extension.
        code = remove_comments(code)
        assert isinstance(code, str)
        self._state["plain_ptyx_code"] = code
        self._state["loaded_extensions"] = extensions
        # Set the seed used for pseudo-random numbers generation.
        # (The seed value is set in the ptyx file using special tag #SEED{}).
        self._state["seed"] = seed
        self._assert_no_comment(code)
        assert "#INCLUDE{" not in code
        assert "#LOAD{" not in code
        assert "#SEED{" not in code
        # Save pTyX code generated by extensions (this is used for debugging,
        # but if needed extensions can also save some data this way using #COMMENT tag).
        # If input file was /path/to/file/myfile.ptyx,
        # plain pTyX code is saved in /path/to/file/.myfile.ptyx.plain-ptyx
        if extensions:
            path = self._state.get("path")
            if path is not None:
                filename = join(dirname(path), ".%s.plain-ptyx" % basename(path))
                with open(filename, "w") as f:
                    f.write(code)

    def generate_syntax_tree(self) -> None:
        code = self._state.get("plain_ptyx_code")
        if code is None:
            raise RuntimeError("Compiler.preparse() must be run first.")
        self._state["syntax_tree"] = self.syntax_tree_generator.generate_tree(code)

    def get_latex(self, **context) -> str:
        """Compile pTyX code and return LaTeX code.

        `Compiler.generate_syntax_tree()` must be run first.
        """
        tree = self._state.get("syntax_tree")
        if tree is None:
            raise RuntimeError("`Compiler.generate_syntax_tree()` must be run first.")
        gen = self.latex_generator
        gen.clear()
        gen.context.update(context)
        seed = self._state["seed"] + gen.NUM
        randfunc.set_seed(seed)
        try:
            gen.parse_node(tree)
        except Exception as e:
            print("\n*** Error occurred while generating code. ***")
            print("This is current compiler state for debugging purpose:")
            print(80 * "-")
            print("... " + "".join(gen.context["PTYX_LATEX"][-10:]))
            print(80 * "-")
            print("")
            if hasattr(e, "msg"):
                print(e.msg)
            raise
        latex = gen.read()
        if "API_VERSION" not in gen.context:
            print("Warning: no API version specified. This may be an old pTyX file.")
        return latex

    def add_new_tags(self, *tags: Tuple[str, Tuple]) -> None:
        """Add ability for extensions to extend syntax, adding new tags."""
        for name, syntax in tags:
            self.syntax_tree_generator.tags[name] = syntax
            self.latex_generator.parser.tags[name] = syntax
        if tags:
            self.syntax_tree_generator.update_tags()
            self.latex_generator.parser.update_tags()

    # TODO: change the signature of parse, to force use of `code` as a keyword argument.
    # Note that this will involve rewriting a lot of tests in both ptyx and ptyx-mcq projects.
    #   parse(self, *, code: str = None, path: Union[Path, str] = None, **context)

    def parse(self, code: str = None, *, path: Union[Path, str] = None, **context) -> str:
        """Convert ptyx code to plain LaTeX in one shot.

        This is mainly used for testing (in unit tests or in interactive mode).

        One may provide either directly the pTyX code, or the path of a pTyX file to be read.

        If both `code` and `path` are provided, the compiler acts as if `code` was the content
        of the pTyX file located at `path`.
        """
        self.reset()
        if path is None and code is None:
            raise ValueError(
                "You must provide either the pTyX code or the path to a pTyX file:\n"
                'compiler.parse(code="...") or compiler.parse(path="/path/to/file.ptyx")'
            )
        if path is not None:
            self.read_file(path)
        if code is not None:
            self.read_code(code)
        self.preparse()
        self.generate_syntax_tree()
        latex = self.get_latex(**context)
        return latex

    @property
    def syntax_tree(self) -> Node:
        return self._state["syntax_tree"]

    @property
    def seed(self) -> Optional[int]:
        return self._state["seed"]

    @property
    def file_path(self) -> Optional[Path]:
        return self._state["path"]

    @property
    def plain_ptyx_code(self) -> Optional[str]:
        return self._state["plain_ptyx_code"]

    @property
    def loaded_extensions(self) -> List[str]:
        return list(self._state["loaded_extensions"].keys())


compiler = Compiler()

parse = compiler.parse



import random
import math

import builtins

import ptyx.randfunc as randfunc
from ptyx.printers import sympy2latex
from ptyx.config import SYMPY_AVAILABLE, NUMPY_AVAILABLE
from ptyx.utilities import latex_verbatim

GLOBAL_CONTEXT = dict()

for fname in [
    "abs",
    "all",
    "any",
    "ascii",
    "bin",
    "callable",
    "chr",
    "divmod",
    "format",
    "getattr",
    "hasattr",
    "hash",
    "hex",
    "id",
    "isinstance",
    "issubclass",
    "iter",
    "len",
    "max",
    "min",
    "next",
    "oct",
    "ord",
    "pow",
    "print",
    "repr",
    "round",
    "sorted",
    "sum",
]:
    GLOBAL_CONTEXT[fname] = getattr(builtins, fname)


if SYMPY_AVAILABLE:
    import sympy

    GLOBAL_CONTEXT["sympy"] = sympy
    GLOBAL_CONTEXT["sympify"] = GLOBAL_CONTEXT["SY"] = sympy.sympify
    # ~ for name in math_list:
    # ~ global_context[name] = getattr(sympy, name)
    exec("from sympy import *", GLOBAL_CONTEXT)
    exec('var("x y")', GLOBAL_CONTEXT)
    # ~ global_context['x'] = sympy.Symbol('x')

if NUMPY_AVAILABLE:
    import numpy

    GLOBAL_CONTEXT["numpy"] = numpy

GLOBAL_CONTEXT["sign"] = lambda x: ("+" if x > 0 else "-")
GLOBAL_CONTEXT["round"] = round
GLOBAL_CONTEXT["min"] = min
GLOBAL_CONTEXT["max"] = max
GLOBAL_CONTEXT["rand"] = GLOBAL_CONTEXT["random"] = random.random
GLOBAL_CONTEXT["ceil"] = GLOBAL_CONTEXT["ceiling"] if sympy is not None else math.ceil
GLOBAL_CONTEXT["float"] = float
GLOBAL_CONTEXT["int"] = int
GLOBAL_CONTEXT["str"] = str

for fname in (
    "randpoint",
    "srandpoint",
    "randint",
    "randbool",
    "srandint",
    "randsign",
    "randfrac",
    "srandfrac",
    "randfloat",
    "srandfloat",
    "randchoice",
    "srandchoice",
    "randmatrix",
    "randpop",
    "shuffle",
    "many",
    "distinct",
    "_print_state",
    "randmaketrans",
):
    GLOBAL_CONTEXT[fname] = getattr(randfunc, fname)

# If a document is compiled several times (to produce different versions of the same document),
# PTYX_NUM is the compilation number (starting from 0).
GLOBAL_CONTEXT["PTYX_NUM"] = 0
GLOBAL_CONTEXT["latex"] = sympy2latex
GLOBAL_CONTEXT["latex_verbatim"] = latex_verbatim



import asyncio

Command = list[str]


async def _run_command(command: Command, shell=False) -> tuple[str, str]:
    if shell:
        proc = await asyncio.create_subprocess_shell(
            " ".join(command), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )
    else:
        proc = await asyncio.create_subprocess_exec(
            *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )
    stdout, stderr = await proc.communicate()
    return stdout.decode().strip(), stderr.decode().strip()


async def _gather(commands: list[Command], shell=False):
    tasks = [asyncio.create_task(_run_command(command, shell)) for command in commands]
    results = await asyncio.gather(*tasks)
    output_list = []
    for result in results:
        output_list.append(result[0])
    return output_list


def run_commands(commands: list[Command], shell=False, max_processes=None) -> list[str]:
    """Run commands asynchronously. The number of processes may be limited using `max_processes`."""
    if max_processes is None:
        max_processes = len(commands)
    packs = [commands[i : i + max_processes] for i in range(0, len(commands), max_processes)]
    return sum((asyncio.run(_gather(pack, shell)) for pack in packs), start=[])



"""
A variant of sympy latex printer, based on wxgeometrie one.
"""

# TODO: Clean-up unused wxgeometrie code !
# TODO: This module needs lot's of cleaning and refactoring.

# try:
#     import sympy
#     from sympy import Basic, Symbol, Integer, Float, I, Mul
#     from sympy.core.core import BasicMeta
#     from sympy.printing.latex import LatexPrinter
# except ImportError:
#     print("Warning: `sympy` library not found !")
#     sympy = None  # type: ignore

import sympy
from sympy import Basic, Symbol, Integer, Float, I, Mul
from sympy.printing.latex import LatexPrinter

from ptyx.config import param


class CustomLatexPrinter(LatexPrinter):
    def __init__(self, settings):
        defaults = {
            "decimales": 18,
            "mat_str": "pmatrix",
            "mat_delim": "",
            "mode": "inline",
            "fold_frac_powers": False,
            "fold_short_frac": False,
        }
        self._default_settings.update(defaults)
        LatexPrinter.__init__(self, settings)

    def _print(self, expr, *args, **kwargs):
        """Change sympy printing algorithm.
        Instead of first searching if the object has a ._latex() method,
        we start by searching if a _print_ObjectClass() exist in the printer.

        This makes it easier to customize sympy objects latex printing.

        So, tries the following concepts to print an expression:
            1. Take the best fitting method defined in the printer.
            2. Let the object print itself if it knows how.
            3. As fall-back use the emptyPrinter method for the printer.
        """
        self._print_level += 1
        try:
            # See if the class of expr is known, or if one of its super
            # classes is known, and use that print function
            for cls in type(expr).__mro__:
                printmethod = "_print_" + cls.__name__
                if hasattr(self, printmethod):
                    return getattr(self, printmethod)(expr, *args, **kwargs)

            # If the printer defines a name for a printing method
            # (Printer.printmethod) and the object knows for itself how it
            # should be printed, use that method.
            if self.printmethod and hasattr(expr, self.printmethod):
                return getattr(expr, self.printmethod)(self, *args, **kwargs)

            # Avoid new (strange) behaviour of sympy 1.7+
            if isinstance(expr, str):
                return expr
            # Unknown object, fall back to the emptyPrinter.
            return self.emptyPrinter(expr)
        finally:
            self._print_level -= 1

    def _print_exp(self, expr, exp=None):
        tex = r"\mathrm{e}^{%s}" % self._print(expr.args[0])
        return self._do_exponent(tex, exp)

    def _print_Exp1(self, expr, exp=None):
        return r"\mathrm{e}"

    def _print_Abs(self, *args, **kw):
        res = LatexPrinter._print_Abs(self, *args, **kw)
        return res.replace(r"\lvert", r"|").replace(r"\rvert", r"|")

    def _print_Pi(self, *args, **kw):
        return r"\pi"

    def _print_ImaginaryUnit(self, expr):
        return r"\mathrm{i}"

    def _print_Function(self, expr, exp=None):
        func = expr.func.__name__

        if hasattr(self, "_print_" + func):
            return getattr(self, "_print_" + func)(expr, exp)

        else:
            if exp is not None:
                name = r"\mathrm{%s}^{%s}" % (func, exp)
            else:
                name = r"\mathrm{%s}" % func
            if len(expr.args) == 1 and isinstance(expr.args[0], (Symbol, Integer)):
                return name + "(" + str(self._print(expr.args[0])) + ")"
            else:
                args = [str(self._print(arg)) for arg in expr.args]
                return name + r"\left(%s\right)" % ",".join(args)

    #    def _print_Float(self, expr):
    #        s = LatexPrinter._print_Float(self, expr)
    #        if s.startswith(r'1.0 \times '):  # sympy 0.7.3
    #            return s[11:]
    #        elif s.startswith(r'1.0 \cdot '):  # sympy 0.7.5
    #            return s[10:]
    #        elif r'\times' not in s:
    #            # Ne pas supprimer un zéro de la puissance !
    #            s = s.rstrip('0').rstrip('.')
    #        return s

    def _print_Infinity(self, expr):
        return r"+\infty"

    def _print_Order(self, expr):
        return r"\mathcal{O}\left(%s\right)" % self._print(expr.args[0])

    def _print_Mul(self, expr):
        args = expr.args
        if args[-1] is I:
            if len(args) == 2 and args[0] == -1:
                return LatexPrinter._print_Mul(self, expr)
            return "%s %s" % (self._print(Mul(*args[:-1])), self._print(I))
        return LatexPrinter._print_Mul(self, expr)
        # return ' '.join(self._print(arg) for arg in expr.args)

    def _print_set(self, expr):
        if expr:
            return r"\left\{%s\right\}" % r"\,;\,".join(self._print(val) for val in expr)
        return r"\emptyset"

    def _print_tuple(self, expr):
        return r"\left(%s\right)" % r",\,".join(self._print(item) for item in expr)

    def _print_log(self, expr, exp=None):
        if len(expr.args) == 1 and isinstance(expr.args[0], (Symbol, Integer)):
            tex = r"\ln(%s)" % self._print(expr.args[0])
        else:
            tex = r"\ln\left(%s\right)" % self._print(expr.args[0])
        return self._do_exponent(tex, exp)

    def _print_function(self, expr):
        return r"\mathrm{Fonction}\, " + expr.__name__

    def doprint(self, expr):
        tex = LatexPrinter.doprint(self, expr)
        return tex.replace(r"\operatorname{", r"\mathrm{")


def custom_latex(expr, **settings):
    """Convert expression to LaTeX code."""
    if sympy is None:
        return str(expr)
    return CustomLatexPrinter(settings).doprint(expr)


def sympy2latex(expr, **flags):
    """Convert a sympy expression to LaTeX code."""
    if flags.get("str"):
        latex = str(expr)
    elif isinstance(expr, float) or (sympy and isinstance(expr, Float)) or flags.get("."):
        # -0.06000000000000001 means probably -0.06 ; that's because
        # floating point arithmetic is not based on decimal numbers, and
        # so some decimal numbers do not have exact internal representation.
        # Python str() handles this better than sympy.latex().
        # Keep 'only' 14 digits (precision rarely  exceed 16 or 17 digits with raw floats,
        # and then decreased with successive operations...)
        latex = format(float(expr), ".14g")
        if latex == "-0":
            latex = "0"
    else:
        latex = custom_latex(expr, mode="plain")

    if isinstance(expr, float) or (sympy and isinstance(expr, Basic)):
        # In french, german... a comma is used as floating point.
        # However, if `float` flag is set, floating point is left unchanged
        # (useful for Tikz for example).
        if not flags.get("."):
            # It would be much better to subclass sympy LaTeX printer
            latex = latex.replace(".", param["floating_point"])

    # TODO: subclass sympy LaTeX printer (cf. mathlib in wxgeometrie)
    latex = latex.replace(r"\operatorname{log}", r"\operatorname{ln}")
    return latex



import re
from math import ceil, floor, isnan, isinf
from typing import Sequence, Any


def round_away_from_zero(val, ndigits=0) -> float:
    """Round using round-away-from-zero strategy for halfway cases.

    Python 3+ implements round-half-even, and Python 2.7 has a random behaviour
    from end user point of view (in fact, result depends on internal
    representation in floating point arithmetic).
    """
    val = float(val)
    if isnan(val) or isinf(val):
        return val
    val *= 10**ndigits
    if val >= 0.0:
        val = floor(val + 0.5)
    else:
        val = ceil(val - 0.5)
    val *= 10 ** (-ndigits)
    return val


def find_closing_bracket(text: str, start: int = 0, brackets: str = "{}", detect_strings=True) -> int:
    """Find the closing bracket, starting from position `start`.

    Note that start have to be a position *after* the opening bracket.

    >>> from ptyx.utilities import find_closing_bracket
    >>> find_closing_bracket('{{hello} world !}', start=1)
    16

    By default, inner strings are handled, so that in "{'}'}", second } will be
    reported as closing bracket, since '}' is seen as an inner string.
    To disable the detection of inner strings, use `detect_strings=False`.

    >>> find_closing_bracket("{'}'}", start=1)
    4
    >>> find_closing_bracket("{'}'}", start=1, detect_strings=False)
    2
    """
    text_beginning = text[start : start + 30]
    # for debugging
    index = 0
    balance = 1
    # None if we're not presently in a string
    # Else, string_type may be ', ''', ", or """
    string_type = None

    open_bracket = brackets[0]
    close_bracket = brackets[1]

    # ', ", { and } are matched.
    # Note that if brackets == '[]', bracket ] must appear first in
    # regular expression ('[]"\'[]' is valid, but '[["\']]' is not).
    reg_str = "[%s\"'%s]" if detect_strings else "[%s%s]"
    reg = re.compile(reg_str % (close_bracket, open_bracket))

    if start:
        text = text[start:]
    while balance:
        m = re.search(reg, text)
        if m is None:
            break

        result = m.group()
        i = m.start()
        if result == open_bracket:
            if string_type is None:
                balance += 1
        elif result == close_bracket:
            if string_type is None:
                balance -= 1

        # Brackets in string should not be recorded...
        # so, we have to detect if we're in a string at the present time.
        elif result in ("'", '"'):
            if string_type is None:
                if text[i:].startswith(3 * result):
                    string_type = 3 * result
                    i += 2
                else:
                    string_type = result
            elif string_type == result:
                string_type = None
            elif string_type == 3 * result:
                if text[i:].startswith(3 * result):
                    string_type = None
                    i += 2

        i += 1  # counting the current character as already scanned text
        index += i
        text = text[i:]

    else:
        return start + index - 1  # last character is the searched bracket :-)

    raise ValueError("ERROR: unbalanced brackets (%s) while scanning %s..." % (balance, repr(text_beginning)))


def advanced_split(
    string: str, separator: str, quotes: str = "\"'", brackets: Sequence[str] = ("()", "[]", "{}")
) -> list[str]:
    """Split string "main_string" smartly, detecting brackets group and inner strings.

    Return a list of strings."""
    if len(separator) != 1:
        raise ValueError("Separator must be a single caracter, not %s." % repr(separator))
    if separator in quotes + "".join(brackets):
        raise ValueError("%s can't be used as separator.")
    # Little optimisation since `not in` is very fast.
    if separator not in string:
        return [string]
    breaks: list[int] = [-1]  # those are the points where the string will be cut
    stack = ["."]  # ROOT
    for i, letter in enumerate(string):
        if letter in quotes:
            if stack[-1] in quotes:
                # We are inside a string.
                if stack[-1] == letter:
                    # Closing string.
                    stack.pop()
            else:
                stack.append(letter)
        elif letter == separator and len(stack) == 1:
            breaks.append(i)
        else:
            start: str
            end: str
            # mypy bug: "Unpacking a string is disallowed"
            for start, end in brackets:  # type: ignore
                if letter == start:
                    stack.append(letter)
                elif letter == end:
                    if stack[-1] != start:
                        raise ValueError("Unbalanced brackets in %s !" % repr(string))
                    stack.pop()
    if len(stack) != 1:
        raise ValueError("Unbalanced brackets in %s !" % repr(string))
    breaks.append(None)  # type: ignore
    # mystring[i:None] returns the end of the string.
    return [string[i + 1 : j] for i, j in zip(breaks[:-1], breaks[1:])]


def _float_me_if_you_can(expr):
    """Convert expr to float if possible, else left it untouched."""
    # noinspection PyBroadException
    try:
        return float(expr)
    except Exception:
        return expr


def numbers_to_floats(expr: Any, integers: bool = False, ndigits: int = None) -> float:
    """Convert all numbers (except integers) to floats inside a sympy expression."""
    import sympy

    if not isinstance(expr, sympy.Basic):
        if isinstance(expr, int) and not integers:
            return expr
        elif ndigits is not None:
            return round_away_from_zero(expr, ndigits)
        else:
            return float(expr)
    for sub in sympy.preorder_traversal(expr):
        sub = sympy.sympify(sub)
        if not sub.has(sympy.Symbol) and (integers or not sub.is_Integer):
            new = sub.evalf()
            if ndigits is not None:
                new = round_away_from_zero(new, ndigits)
            expr = expr.subs(sub, new)
    return expr


def term_color(string, color, **kw):
    """On Linux, format string for terminal printing.

    Available keywords: bold, dim, italic, underline and hightlight.

    >>> term_color('hello world !', 'blue', bold=True, underline=True)
    '\x1b[4;1;34mhello world !\x1b[0m'
    """
    colors = {
        "gray": 30,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "blue": 34,
        "purple": 35,
        "cyan": 36,
        "white": 37,
    }
    styles = {"bold": 1, "dim": 2, "italic": 3, "underline": 4, "highlight": 7}
    if color not in colors:
        raise KeyError("Color %s is unknown. Available colors: %s." % (repr(color), list(colors.keys())))
    formatting = []
    for style, code in styles.items():
        appply_style = kw.get(style)
        if appply_style is not None:
            formatting.append(str(code) if appply_style else str(20 + code))
    formatting.append(str(colors[color]))
    return "\033[%sm%s\033[0m" % (";".join(formatting), string)


def latex_verbatim(s: str) -> str:
    """Try to emulate verbatim (which is not allowed inside a macro argument in LaTeX).

    LaTeX special characters are escaped, so that the string is displayed as it.
    A fixed-width font is used too.

        >>> latex_verbatim("\\emph{$a^2+b_i$}")
        '\\texttt{\\textbackslash{}emph\\{\\$a\\textasciicircum{}2+b\\_i\\$\\}}'
    """
    # Strip the first \n, to avoid beginning with a \linebreak.
    if s.startswith("\n"):
        s = s[1:]
    # Replace \ first !
    s = s.replace("\\", r"\textbackslash<!ø5P3C14Lø?>")
    s = s.replace("~", r"\textasciitilde<!ø5P3C14Lø?>")
    s = s.replace("^", r"\textasciicircum<!ø5P3C14Lø?>")
    s = s.replace("'", r"\textquotesingle<!ø5P3C14Lø?>")
    for char in "#$%&_{}":
        s = s.replace(char, rf"\{char}")
    s = s.replace("<!ø5P3C14Lø?>", "{}")
    # Use \phantom{} after \linebreak to preserve spaces at the beginning of the line.
    s = s.replace("\n", "\\linebreak\\phantom{}")
    s = s.replace("\t", 4 * " ")
    s = s.replace(" ", "~")
    return rf"\texttt{{{s}}}"



from os.path import join, dirname

from ptyx.extensions.questions import main


def test_question():
    with open(join(dirname(__file__), "in.txt")) as f:
        questions_text = f.read()

    with open(join(dirname(__file__), "out.ptyx")) as f:
        expected_ptyx_code = f.read()

    assert main(questions_text, None) == expected_ptyx_code



import itertools
import multiprocessing
import os
import re
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Optional, Dict, Iterable, List, Sequence, Tuple, Union

import fitz

from ptyx.config import param, CPU_PHYSICAL_CORES
from ptyx.latex_generator import compiler


def append_suffix(path: Path, suffix) -> Path:
    """>>> append_suffix(Path("/home/user/file"), "-corr")
    Path("/home/user/file-corr")
    """
    return path.with_name(path.name + suffix)


class _LoggedStream(object):
    """Add logging to a data stream, like stdout or stderr.

    * `logfile` is a file already opened in appending mode ;
    * `default` is default output (`sys.stdout` or `sys.stderr`).
    """

    def __init__(self, logfile, default):
        self.logfile = logfile
        self.default = default

    def write(self, s):
        self.default.write(s)
        self.logfile.write(s)

    def flush(self):
        self.default.flush()
        self.logfile.flush()


class _DevNull(object):
    def write(self, *_):
        pass

    close = flush = write


class Logging(object):
    """Context manager. All output (sys.stdout/stderr) will be logged in a file.

    Note this logging occurs in addition to standard output, which is not suppressed.
    """

    def __init__(self, logfile_name: Path | str | None = None):
        self.logfile = open(logfile_name, "a") if logfile_name else _DevNull()

    def __enter__(self):
        self.previous = {"stdout": sys.stdout, "stderr": sys.stderr}
        sys.stdout = _LoggedStream(self.logfile, sys.stdout)
        sys.stderr = _LoggedStream(self.logfile, sys.stderr)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        assert isinstance(sys.stdout, _LoggedStream)
        assert isinstance(sys.stderr, _LoggedStream)
        sys.stdout = sys.stdout.default
        sys.stderr = sys.stderr.default

        self.logfile.close()


# def execute(string: str, quiet=False) -> str:
#     out = subprocess.Popen(string, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).stdout
#     if out is not None:
#         encoding = locale.getpreferredencoding(False)
#         output = out.read().decode(encoding, errors="replace")
#         sys.stdout.write(output)
#         out.close()
#     else:
#         output = ""
#     if not quiet:
#         print(f"Command '{string}' executed.")
#     return output


def execute(command: str) -> str:
    """Execute command in shell."""
    out = subprocess.Popen(
        command, shell=True, encoding="utf8", stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    ).stdout
    return out.read() if out is not None else ""


def make_files(
    input_name: Path,
    correction: bool = False,
    fixed_number_of_pages: bool = False,
    pages: Optional[int] = None,
    context: Optional[dict] = None,
    quiet: Optional[bool] = None,
    remove: Optional[bool] = None,
    _nums: Iterable[int] = None,
    **options,
) -> Tuple[Path, List[int]]:
    """Generate the tex and pdf files.

    - `correction`: if True, include the solutions of the exercises
    - `fixed_number_of_pages`: if True, all pdf files must have the same number of pages
    - `pages`: keep only pdf documents respecting this number of pages contraint
    - `context`: parameters to be passed to the LaTeX generator.
    - `quiet`: if True, turn off debugging information
    - `remove`: if True, remove `.compile` folder after successful compilation.
    """

    target = options.get("number_of_documents", param["total"])
    # `_nums` is used when generating the answers of the quiz.
    # In the first pass, when generating the quizzes, some numbers may
    # have been skipped (because they don't satisfy the pages number constraint).
    if _nums is not None:
        assert correction
        _nums = list(_nums)  # make a copy
        target = len(_nums)
    if context is None:
        context = {"PTYX_WITH_ANSWERS": correction}
    formats: list[str] = options.get("formats", param["default_formats"].split("+"))  # type: ignore
    cpu_cores: Optional[int] = options.get("cpu_cores")  # type: ignore

    # Create an empty `.compile/{input_name}` subfolder.
    compilation_dir = input_name.parent / ".compile" / input_name.stem
    if not correction and compilation_dir.is_dir():
        shutil.rmtree(compilation_dir)
    compilation_dir.mkdir(parents=True, exist_ok=True)

    # Set output base name
    output_basename = compilation_dir / input_name.stem
    if input_name.suffix != ".ptyx":
        # Avoid potential name conflict between input and output.
        output_basename = append_suffix(output_basename, "_")
    if correction:
        output_basename = append_suffix(output_basename, "-corr")

    # Information to collect
    compilation_info: Dict[int, Path] = {}
    # nums: List[int] = []
    # filenames: List[Path] = []
    pages_per_document: Dict[int, Dict[int, Path]] = {}

    # Compilation number, used to initialize random numbers' generator.
    num = options.get("start", 1) - 1
    assert target is not None
    while len(compilation_info) < target:
        # ------------------------
        # Generate the LaTeX files
        # ------------------------
        # (Note that the actual number of generated files may be more than that,
        # because by default we aim to have documents with the same number of pages.)
        latex_files: list[Path] = []
        nums: list[int] = []
        filenames: list[Path] = []
        for _ in range(target - len(compilation_info)):
            # 1. Generate context.
            num += 1
            if correction and _nums is not None:
                # Overwrite default numeration, since correction numeration must match first pass.
                num = _nums.pop(0)
            context.update(PTYX_NUM=num)
            nums.append(num)
            filename = append_suffix(output_basename, f"-{num}") if target > 1 else output_basename
            filenames.append(filename)
            # 2. Compile to LaTeX.
            latex_files.append(generate_latex(filename, context))

        # --------------------------------
        # Compile to pdf using parallelism
        # --------------------------------
        # Use only the physical cores, not the virtual ones !
        # with Pool(CPU_PHYSICAL_CORES) as pool:
        #     infos_list = pool.starmap(make_file, tasks)

        cpu_cores_to_use = cpu_cores if cpu_cores else min(CPU_PHYSICAL_CORES, len(latex_files))
        args = [(path, None, quiet) for path in latex_files]

        if cpu_cores_to_use > 1:
            with multiprocessing.get_context("forkserver").Pool(cpu_cores_to_use) as pool:
                pages_numbers: list[int] = pool.starmap(compile_latex, args)
        else:
            pages_numbers = list(itertools.starmap(compile_latex, args))

        # ---------------
        # Analyze results
        # ---------------
        for pdf_pages_number, num, filename in zip(pages_numbers, nums, filenames):
            if not correction:
                # 3. Test if the new generated file satisfies all options constraints.
                if fixed_number_of_pages:
                    if pages is None:
                        # This is a bit subtle. We want all compiled documents to have
                        # the same pages number, yet we don't want to set it manually.
                        # So, we compile documents and memorize their size.
                        # We'll group compilation results by the length of the resulting document.
                        # We'll keep one dictionary {document id: Path}  for each size of document.
                        # At each loop, if the compiled document is of size n,
                        # we update the dictionary of all the n-sized documents with the document ID
                        # and its path.
                        # Before beginning the next loop, the dictionary size will be checked,
                        # to see if it contains enough documents.
                        # (There's no need to have a look on the others dicts, since they haven't change...)
                        compilation_info = pages_per_document.setdefault(pdf_pages_number, {})
                    elif pages != pdf_pages_number:
                        # Pages number is set manually, and don't match.
                        print(f"Warning: skipping {filename} (incorrect page number) !")
                        continue

            compilation_info[num] = filename
            num += 1

    assert len(compilation_info) == target
    filenames = list(compilation_info.values())

    # Join different versions in a single pdf, and compress if asked to do so.
    join_files(output_basename, filenames, **options)

    if options.get("generate_batch_for_windows_printing"):
        bat_file_name = input_name.parent / ("print_corr.bat" if correction else "print.bat")
        with open(bat_file_name, "w") as bat_file:
            bat_file.write(
                param["win_print_command"]
                + " ".join('"%s.pdf"' % os.path.basename(f) for f in filenames)  # type: ignore
            )

    # Copy tex/pdf file to parent directory.
    for ext in formats:
        assert ext[0] != "."
        ext = f".{ext}"
        name = output_basename.with_suffix(ext)
        if name.is_file():
            # There is only one file (only one document was generated,
            # or they were several documents, but they were joined into a single document).
            shutil.copy(name, input_name.parent)
        elif options.get("names_list"):
            # Rename files according to the given names' list.
            names = options["names_list"]
            assert len(names) == len(filenames)
            for filename, stem in zip(filenames, names):
                new_name = filename.with_stem(stem).name
                shutil.copy(filename.with_suffix(ext), input_name.parent / new_name)
        else:
            # Copy files without changing names.
            for filename in filenames:
                shutil.copy(filename.with_suffix(ext), input_name.parent)

    # Remove `.compile` folder if asked to.
    if remove:
        shutil.rmtree(compilation_dir)

    return output_basename, list(compilation_info.keys())


def generate_latex(
    output_name: Path,
    context: Optional[Dict] = None,
    log=True,
) -> Path:
    """Generate latex from ptyx source file."""
    if log:
        # Output is redirected to a `.log` file.
        logfile: Optional[Path] = append_suffix(output_name, "-python.log")
        print("\nLog file:", logfile, "\n")
    else:
        logfile = None

    with Logging(logfile):
        if context is None:
            context = {}

        context.setdefault("PTYX_NUM", 1)
        latex = compiler.get_latex(**context)

        texfile_path = output_name.with_suffix(".tex")
        with open(texfile_path, "w") as texfile:
            texfile.write(latex)
        return texfile_path


def make_file(
    output_name: Path,
    context: Optional[Dict] = None,
    quiet: Optional[bool] = None,
) -> int:
    """Generate latex and/or pdf file from ptyx source file."""
    return compile_latex(generate_latex(output_name, context), quiet=quiet)


def _print_latex_errors(out: str, filename: Path):
    print(f"File {filename} compiled.")
    for line in out.split("\n"):
        if line.startswith("!"):
            print(line)
    print(f"Full log written on {filename.with_suffix('.log')}.")


def compile_latex(filename: Path, dest: Optional[Path] = None, quiet: Optional[bool] = False) -> int:
    """Compile the latex file and return the number of pages of the pdf (or -1 if not found)."""
    command = _build_command(filename, dest, quiet)
    out = execute(command)
    _print_latex_errors(out, filename)
    # Run command twice if references were found.
    if "Rerun to get cross-references right." in out or "There were undefined references." in out:
        # ~ input('- run again -')
        out = execute(command)
        _print_latex_errors(out, filename)
    return _extract_page_number(out)


def _build_command(filename: Path, dest: Optional[Path] = None, quiet: Optional[bool] = False) -> str:
    """Generate the command used to compile the LaTeX file."""
    # By default, pdflatex use current directory as destination folder.
    # However, much of the time, we want destination folder to be the one
    # where the tex file was found.
    if dest is None:
        dest = filename.parent

    command: str = param["quiet_tex_command"] if quiet else param["tex_command"]  # type: ignore
    command += f' -output-directory "{dest}" "{filename}"'
    return command


def _extract_page_number(pdflatex_log: str) -> int:
    """Return the number of pages of the pdf generated, or -1 if it was not found."""
    i = pdflatex_log.find("Output written on ")
    if i == -1:
        return -1
    pattern = r"Output written on .+ \(([0-9]+) pages, [0-9]+ bytes\)\."
    # Line breaks may occur anywhere in the log after the file path,
    # so using re.DOTALL flag is not enough, we have to manually remove all `\n`.
    m = re.search(pattern, pdflatex_log[i:].replace("\n", ""))
    return int(m.group(1)) if m is not None else -1


def _join_pdf_files(output_basename: Path | str, pdfnames: Sequence[Path | str]) -> None:
    """Join all the generated pdf files into one file."""
    # pdf: Document
    if len(pdfnames) == 0:
        print("Warning: no PDF files to join.")
        return
    with fitz.Document() as pdf:
        for pdfname in pdfnames:
            # f: Document
            with fitz.Document(pdfname) as f:
                pdf.insert_pdf(f)
        pdf.save(output_basename)


def join_files(
    filename: Path,
    pdfnames: Sequence[Union[Path, str]],
    seed_file_name=None,
    **options,
):
    """Join different versions in a single pdf, then compress it if asked to do so."""
    # TODO: use pathlib.Path instead
    pdf_name = str(filename) + ".pdf"
    number = len(pdfnames)

    if options.get("compress") or options.get("cat"):
        # Nota: don't exclude the case `number == 1`,
        # since the following actions rename file,
        # so excluding the case `number == 1` would break autoqcm scan for example.
        # pdftk and ghostscript must be installed.
        pdfnames = [str(filename) + ".pdf" for filename in pdfnames]

        if len(pdfnames) > 1:
            _join_pdf_files(f"{filename}.pdf", pdfnames)
        if options.get("remove_all"):
            for name in pdfnames:
                os.remove(name)
        if options.get("compress"):
            temp_dir = tempfile.mkdtemp()
            compressed_pdf_name = os.path.join(temp_dir, "compresse.pdf")
            command = f"""command pdftops \
                -paper match \
                -nocrop \
                -noshrink \
                -nocenter \
                -level3 \
                -q \
                "{pdf_name}" - \
                | command ps2pdf14 \
                -dEmbedAllFonts=true \
                -dUseFlateCompression=true \
                -dProcessColorModel=/DeviceCMYK \
                -dConvertCMYKImagesToRGB=false \
                -dOptimize=true \
                -dPDFSETTINGS=/prepress \
                - "{compressed_pdf_name}" """
            os.system(command)
            old_size = os.path.getsize(pdf_name)
            new_size = os.path.getsize(compressed_pdf_name)
            if new_size < old_size:
                shutil.copyfile(compressed_pdf_name, pdf_name)
                print(f"Compression ratio: {old_size / new_size:.2f}")
            else:
                print("Warning: compression failed.")
            if seed_file_name is not None:
                temp_dir = tempfile.mkdtemp()
                pdf_with_seed = os.path.join(temp_dir, "with_seed.pdf")
                execute(f'pdftk "{pdf_name}" attach_files "{seed_file_name}" output "{pdf_with_seed}"')
                shutil.copyfile(pdf_with_seed, pdf_name)
        if number > 1:
            print(f"{len(pdfnames)} files merged.")

    if options.get("reorder_pages"):
        # Use pdftk to detect how many pages has the pdf document.
        n = int(execute(f"pdftk {pdf_name} dump_data output | grep -i NumberOfPages:").strip().split()[-1])
        mode = options.get("reorder_pages")
        if mode == "brochure":
            if n % 4:
                raise RuntimeError(f"Page number is {n}, but must be a multiple of 4.")
            order = []
            for i in range(int(n / 4)):
                order.extend([2 * i + 1, 2 * i + 2, n - 2 * i - 1, n - 2 * i])
        elif mode == "brochure-reversed":
            if n % 4:
                raise RuntimeError(f"Page number is {n}, but must be a multiple of 4.")
            order = n * [0]
            for i in range(int(n / 4)):
                order[2 * i] = 4 * i + 1
                order[2 * i + 1] = 4 * i + 2
                order[n - 2 * i - 2] = 4 * i + 3
                order[n - 2 * i - 1] = 4 * i + 4
        else:
            raise NameError(f"Unknown mode {mode} for option --reorder-pages !")
        # monfichier.pdf -> monfichier-brochure.pdf
        new_name = "%s-%s.pdf" % (pdf_name[: pdf_name.index(".")], mode)
        execute("pdftk %s cat %s output %s" % (pdf_name, " ".join(str(i) for i in order), new_name))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 23 15:09:44 2021

@author: Nicolas Pourcelot
"""

import re
from typing import Tuple, Optional, List, Dict, Union, TypeVar, Iterable, Set, Any

from ptyx.utilities import find_closing_bracket, term_color

Tag = str
TagSyntax = Tuple[int, int, Optional[List[str]]]
NodeChild = Union[str, "Node"]
S = TypeVar("S")
T = TypeVar("T", bound=NodeChild)


class Node:
    """A node.

    `name` is either the tag name, if the node corresponds
    to a tag's content, or the argument number, if the node corresponds
    to a tag's argument."""

    def __init__(self, name: Union[str, int]):
        self.parent: Optional[Node] = None
        self.name = name
        self.options: Optional[str] = None
        self.children: List[NodeChild] = []

    def __repr__(self):
        return f"<Node {self.name} at {hex(id(self))}>"

    def add_child(self, child: T) -> Optional[T]:
        if not child:
            return None
        self.children.append(child)
        if isinstance(child, Node):
            child.parent = self
        return child

    # @property
    # def content(self):
    #   if len(self.children) != 1:
    #       raise ValueError, "Ambiguous: node has several subnodes."
    #   return self.children[0]

    def arg(self, i: int) -> Any:
        """Return argument number i content, which may be a Node, or some text (str)."""
        child = self.children[i]
        if getattr(child, "name", None) != i:
            raise ValueError(f"Incorrect argument number for node {child!r}.")
        assert isinstance(child, Node), repr(child)
        children_number = len(child.children)
        if children_number > 1:
            raise ValueError(
                f"Don't use pTyX code inside {self.name} argument number {i + 1}.\n"
                "------------------\n"
                f"{child.display()}\n"
                "------------------\n"
            )
        if children_number == 0:
            if self.name == "EVAL":
                # EVAL isn't a real tag name: if a variable `#myvar` is found
                # somewhere, it is parsed as an `#EVAL` tag with `myvar` as argument.
                # So, a lonely `#` is parsed as an `#EVAL` with no argument at all.
                raise ValueError("Error! There is a lonely '#' somewhere !")
            print("Warning: %s argument number %s is empty." % (self.name, i + 1))
            return ""

        return child.children[0]

    def display(self, color=True, indent=0, raw=False) -> str:
        texts = ["%s+ Node %s" % (indent * " ", self._format(self.name, color))]
        for child in self.children:
            if isinstance(child, Node):
                texts.append(child.display(color, indent + 2, raw=raw))
            else:
                assert isinstance(child, str)
                text = repr(child)
                if not raw:
                    if len(text) > 30:
                        text = text[:25] + " [...]'"
                if color:
                    text = term_color(text, "green")
                texts.append("%s  - text: %s" % (indent * " ", text))
        return "\n".join(texts)

    def as_text(self, skipped_children: Iterable[int] = ()) -> str:
        content = []
        for i, child in enumerate(self.children):
            if i in skipped_children:
                continue
            if isinstance(child, Node):
                content.append(child.as_text())
            else:
                assert isinstance(child, str)
                content.append(child)
        return "".join(content)

    @staticmethod
    def _format(val: object, color: bool) -> str:
        if not color:
            return str(val)
        if isinstance(val, str):
            return term_color(val, "yellow")
        if isinstance(val, int):
            return term_color(str(val), "blue")
        return str(val)


class SyntaxTreeGenerator:
    # For each tag, indicate:
    #   1. The number of interpreted arguments (arguments that contain code).
    #      Those arguments will not be parsed.
    #      (This is used most of the time for python code).
    #   2. The number of raw arguments.
    #      Those arguments contain raw text, which need parsing.
    #   3. If the tag opens a block, a list of all the tags closing the block,
    #      else `None`.
    #
    # Notice that by default, the tag closing the block will not be consumed.
    # This means that the same tag will be parsed again to open or close another block.
    # To consume the closing tag, prefix the tag name with the '@' symbol.
    # This is usually the wished behaviour for #END tag.
    #
    # Distinction between code arguments and raw arguments must be done because
    # in raw arguments, there should be no detection of inner strings:
    # in {$f'(x)$}, the ' must not be interpreted as an opening string, so closing
    # bracket is the one following the $.
    # By contrast, in code arguments, inner strings should be detected:
    # in {val=="}"}, the bracket closing the tag is the second `}`, not the first one !

    tags: Dict[Tag, TagSyntax] = {
        "ANS": (0, 0, ["@END"]),
        "ANSWER": (0, 1, None),
        "APART": (0, 0, ["@END", "@END_APART"]),
        "ASK": (0, 0, ["@END"]),
        "ASK_ONLY": (0, 0, ["@END"]),
        "ASSERT": (1, 0, None),
        # Do *NOT* consume #END tag, which must be used to end #CONDITIONAL_BLOCK.
        "CASE": (1, 0, ["CASE", "ELSE", "END", "END_CASE"]),
        "COMMENT": (0, 0, ["@END"]),
        # CONDITIONAL_BLOCK isn't a real tag, but is used to enclose
        # a #CASE{...}...#CASE{...}...#END block, or an #IF{...}...#ELIF{...}...#END block.
        "CONDITIONAL_BLOCK": (0, 0, ["@END", "@END_IF"]),
        "DEBUG": (0, 0, None),
        "EVAL": (1, 0, None),
        # ENUM indicates the start of an enumeration.
        # It does nothing by itself, but is used by some extensions.
        "ENUM": (0, 0, ["@END"]),
        # Do *NOT* consume #END tag, which must be used to end #CONDITIONAL_BLOCK.
        "IF": (1, 0, ["ELIF", "ELSE", "END", "END_IF"]),
        # Do *NOT* consume #END tag, which must be used to end #CONDITIONAL_BLOCK.
        "ELIF": (1, 0, ["ELIF", "ELSE", "END", "END_IF"]),
        # Do *NOT* consume #END tag, which must be used to end #CONDITIONAL_BLOCK.
        "ELSE": (0, 0, ["END", "END_IF"]),
        "IMPORT": (1, 0, None),
        "LOAD": (1, 0, None),
        "FREEZE_RANDOM_STATE": (0, 0, []),
        "IFNUM": (1, 1, None),
        "INCLUDE": (1, 0, None),
        "CALL": (0, 1, None),
        "MACRO": (0, 1, ["@END", "@END_MACRO"]),
        "PICK": (0, 0, ["@END", "@END_PICK"]),
        "PRINT": (1, 0, None),
        "PTYX_VERSION": (0, 1, None),
        "PYTHON": (0, 0, ["@END"]),
        "QUESTION": (0, 1, None),
        # ROOT isn't a real tag, and is never closed.
        "ROOT": (0, 0, []),
        "SEED": (1, 0, None),
        "SHUFFLE": (0, 0, ["@END", "@END_SHUFFLE"]),
        # Do *NOT* consume #END tag, which must be used to end #SHUFFLE block.
        "ITEM": (0, 0, ["ITEM", "END", "END_SHUFFLE", "END_PICK"]),
        "SIGN": (0, 0, None),
        "SYMPY": (0, 0, ["@END"]),
        "TEST": (1, 2, None),
        "VERBATIM": (0, 0, ["@END", "@END_VERBATIM"]),
        "-": (0, 0, None),
        "+": (0, 0, None),
        "*": (0, 0, None),
        "=": (0, 0, None),
        "?": (0, 0, None),
        "#": (0, 0, None),
    }

    # NOTE: Should all tags starting with END_TAG automatically close TAG ?
    # (Should this be a syntax feature ?
    # It sounds nice, but how should we deal with the `@` then ?).

    def __init__(self):
        # Add ability to update the set of closing tags for instances of
        # SyntaxTreeGenerator.
        # It is used by extensions to define new closing tags,
        # by calling `Compiler.add_new_tag()`.
        self._found_tags: Set[Tag] = set()
        self.reset()

    def reset(self):
        """Full reset."""
        self.tags = dict(self.tags)
        self.update_tags()
        self.syntax_tree = None

    @staticmethod
    def only_closing(tag):
        """Return `True` if tag is only a closing tag, `False` else."""
        return tag == "END" or tag.startswith("END_")

    def update_tags(self):
        """Automatically add closing tags, then generate sorted list."""
        missing = set()
        for name, syntax in self.tags.items():
            closing_tags = syntax[2]
            if closing_tags is None:
                continue
            for tag in closing_tags:
                tag = tag.lstrip("@")
                if tag not in self.tags:
                    missing.add(tag)

        for name in missing:
            self.tags[name] = (0, 0, None)

        # Tags sorted by length (longer first).
        # This is used for matching tests.
        self.sorted_tags = sorted(self.tags, key=len, reverse=True)

    @staticmethod
    def remove_comments(text: str) -> str:
        # If the comment is at the end of a line, don't remove the end of line.
        # However, if the full line is a comment, remove the end of line (\n).
        text = re.sub("( # .+)|(^# .+\n)", "", text, flags=re.MULTILINE)
        return text

    def generate_tree(self, text):
        """Pre-parse pTyX code and generate a syntax tree.

        :param text: some pTyX code.
        :type text: string

        .. note:: To access generated syntax tree, use `.syntax_tree` attribute.
        """
        # Now, we will parse Ptyx code to generate a syntax tree.
        self._found_tags = set()
        self.syntax_tree = Node("ROOT")
        # Remove all comments from text.
        text = self.remove_comments(text)
        self._generate_tree(self.syntax_tree, text)
        self.syntax_tree.tags = self._found_tags
        return self.syntax_tree

    def _generate_tree(self, node, text):
        """Parse `text`, then add corresponding content to `node`."""
        position = 0
        update_last_position = True
        node._closing_tags = []

        while True:
            # --------------
            # Find next tag.
            # --------------
            # A tag starts with '#'.

            # last_position should not be updated if false positives
            # were encountered (ie. #1, #2 in \newcommand{}...).
            if update_last_position:
                last_position = position
            else:
                update_last_position = True
            position = tag_position = text.find("#", position)
            if position == -1:
                # No tag anymore.
                break
            position += 1
            # Is this a known tag ?
            for tag in self.sorted_tags:
                if text[position:].startswith(tag):
                    # Mmm, this begins like a known tag...
                    # In fact, it will really match a known tag if one of the following occurs:
                    # - next character is not alphanumeric ('#IF{' for example).
                    # - tag is not alphanumeric ('#*' tag for example).
                    if not tag[-1].replace("_", "a").isalnum():
                        # Tag is not alphanumeric, so no confusion with a variable name can occur.
                        # -> yes, a known tag found !
                        position += len(tag)
                        break
                    if position + len(tag) == len(text):
                        # There is no next character (last text character reached).
                        # -> yes, a known tag found !
                        position += len(tag)
                        break
                    next_character = text[position + len(tag)]
                    if not (next_character == "_" or next_character.isalnum()):
                        # Next character is not alphanumeric
                        # -> yes, a known tag found !
                        position += len(tag)
                        break
                    # -> sorry, try again.
            else:
                if position >= len(text) or text[position].isdigit() or text[position] == " ":
                    # This not a tag: LaTeX uses #1, #2, #3 as \newcommand{} parameters.
                    # This may also be a simple \# .
                    # Pretend nothing happened.
                    update_last_position = False
                    continue
                else:
                    # This is not a known tag name.
                    # Default tag name is EVAL.
                    # Notably:
                    # - any variable (like #a)
                    # - any expression (like #{a+7})
                    # will result in an #EVAL tag.
                    tag = "EVAL"

            # ------------------------
            # Deal with new found tag.
            # ------------------------
            # Syntax tree maintains a record of all tags found during parsing.
            # (This tags set can be later accessed through its .tags attribute.)
            self._found_tags.add(tag)

            # Add text found before this tag to the syntax tree.
            # --------------------------------------------------

            remove_trailing_newline = self.tags[tag][2] is not None or self.only_closing(tag)
            if remove_trailing_newline:
                # Remove new line and spaces *before* #IF, #ELSE, ... tags.
                # This is more convenient, since two successive \n
                # induce a new paragraph in LaTeX.
                # So, something like this
                #   [some text here]
                #   #IF{delta>0}
                #   [some text there]
                #   #END
                # would automatically result in two paragraphs else.
                i = max(text.rfind("\n", None, tag_position), 0)
                if text[i:tag_position].isspace():
                    node.add_child(text[last_position:i])
                else:
                    node.add_child(text[last_position:tag_position])
            else:
                node.add_child(text[last_position:tag_position])

            # Enclose "CASE ... CASE ... ELSE ... END" or "IF ... ELIF ... ELSE ... END"
            # inside a CONDITIONAL_BLOCK node.
            # --------------------------------------------------------------------------
            #
            # The most subtle part while parsing pTyX code is to distinguish
            # between "#CASE{0}...#CASE{1}...#ELSE...#END"
            # and "#CASE{0}...#END#CASE{1}...#ELSE...#END".
            # This distinction is important because, if NUM==0, the ELSE clause
            # must be executed in the 2nde version, but not in the 1st one.
            #
            # This is one of the reasons why CASE nodes are enclosed inside a CONDITIONAL_BLOCK.
            # So, first version must result in only one CONDITIONAL_BLOCK,
            # while 2nd version must result in 2 CONDITIONAL_BLOCKs.
            #
            # The rule is actually quite simple: a #CASE tag must open a new CONDITIONAL_BLOCK
            # only if previous opened node wasn't a #CASE node.
            #
            # Note that for #IF blocks, there is no such difficulty,
            # because an #IF tag always opens a new CONDITIONAL_BLOCK.
            if (tag == "CASE" and node.name != "CASE") or tag == "IF":
                node = node.add_child(Node("CONDITIONAL_BLOCK"))
                node._closing_tags = self.tags["CONDITIONAL_BLOCK"][2]

            # Detect if this tag is actually closing a node.
            # ----------------------------------------------
            while tag in node._closing_tags:
                # Close node, but don't consume tag.
                node = node.parent
            if "@" + tag in node._closing_tags:
                # Close node and consume tag.
                node = node.parent
                continue

            # Special case : don't pre-parse #PYTHON ... #END content.
            # ----------------------------------------------------
            if tag == "PYTHON":
                end = text.index("#END", position)
                # Create and enter new node.
                node = node.add_child(Node(tag))
                # Some specific parsing is done however:
                # a line starting with `%` will be interpreted as a comment.
                # This makes the code a bit more readable, since `%` is already used
                # for comments in LateX code.
                _text = text[position:end]
                _text = re.sub(r"^\s*%", "#", _text, flags=re.MULTILINE)
                node.add_child(_text)
                node = node.parent
                position = end + 4

            # General case
            # ------------
            # Exclude #END and all closing tags, since they're not true tags.
            # (Their only purpose is to close a block, #END doesn't correspond to any command).
            elif not self.only_closing(tag):
                # Create and enter new node.
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~
                node = node.add_child(Node(tag))
                # Detect command optional argument.
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                try:
                    # - Tolerate spaces before bracket.
                    tmp_pos = position
                    while text[tmp_pos].isspace():
                        tmp_pos += 1
                        if text[tmp_pos] == "[":
                            position = tmp_pos
                    # - Handle optional argument.
                    if text[position] == "[":
                        position += 1
                        end = find_closing_bracket(text, position, brackets="[]")
                        node.options = text[position:end]
                        position = end + 1
                except IndexError:
                    # Don't raise error, since argument is optional.
                    pass

                # Detect command arguments.
                # ~~~~~~~~~~~~~~~~~~~~~~~~~
                # Each argument becomes a node with its number as name.
                try:
                    code_args_number, raw_args_number, closing_tags = self.tags[node.name]
                except ValueError:
                    raise RuntimeError("Tag %s is not correctly defined." % node.name)
                for i in range(code_args_number + raw_args_number):
                    try:
                        # - Tolerate spaces before bracket.
                        while text[position].isspace():
                            position += 1
                        # - Handle argument.
                        if text[position] == "{":
                            position += 1
                            # Detect inner strings for arguments containing code,
                            # but not for arguments containing raw text.
                            end = find_closing_bracket(
                                text,
                                position,
                                brackets="{}",
                                detect_strings=(i < code_args_number),
                            )
                            new_pos = end + 1
                        else:
                            end = position
                            while end < len(text) and text[end].isalnum():
                                end += 1
                            new_pos = end
                    except IndexError:
                        raise RuntimeError("Missing argument for tag %s !" % tag)
                    # Each argument of a command is a node itself.
                    # Nodes corresponding to arguments have no name,
                    # but are numbered instead.
                    arg = node.add_child(Node(i))
                    self._generate_tree(arg, text[position:end])
                    position = new_pos

                # if remove_trailing_newline:
                #   # Remove new line and spaces *after* #IF, #ELSE, ... tags.
                #   # This is more convenient, since two successive \n
                #   # induce a new paragraph in LaTeX.
                #   # So, something like this
                #   #   [some text here]
                #   #   #IF{delta>0}
                #   #   #IF{a>0}
                #   #   [some text there]
                #   #   #END
                #   #   #END
                #   # would automatically result in two paragraphs else.
                #   try:
                #       i = text.index('\n', position) + 1
                #       if text[position:i].isspace():
                #           position = i
                #   except ValueError:
                #       pass

                # Close node if needed.
                # ~~~~~~~~~~~~~~~~~~~~~~~
                if closing_tags is None:
                    # Close node (tag is self-closing).
                    node = node.parent
                else:
                    # Store node closing tags for fast access later.
                    node._closing_tags = closing_tags

        node.add_child(text[last_position:])



# mypy: ignore-errors

import os
import sys
import importlib.util

import psutil

try:
    from ptyx.custom_config import param as custom_param
except ImportError:
    custom_param = {}

# <default_configuration>
param = {
    "total": 1,
    "formats": ["pdf", "tex", "tex+pdf", "pdf+tex"],
    "default_formats": "pdf",
    "tex_command": "pdflatex -interaction=nonstopmode --shell-escape --enable-write18",
    "quiet_tex_command": "pdflatex -interaction=batchmode --shell-escape --enable-write18",
    "sympy_is_default": True,
    "import_paths": [],
    "debug": False,
    "floating_point": ",",
    "win_print_command": "SumatraPDF.exe -print-dialog -silent -print-to-default ",
}
# </default_configuration>

# Update parameters using `custom_config.py` param, if any.
param.update(custom_param)


for path in param["import_paths"]:
    path = os.path.normpath(os.path.expanduser(path))
    sys.path.insert(0, path)


SYMPY_AVAILABLE = importlib.util.find_spec("sympy") is not None
NUMPY_AVAILABLE = importlib.util.find_spec("numpy") is not None

if not SYMPY_AVAILABLE:
    print("** ERROR: sympy not found ! **")
    param["sympy_is_default"] = False

if not NUMPY_AVAILABLE:
    print("WARNING: numpy not found.")

try:
    CPU_PHYSICAL_CORES = psutil.cpu_count(logical=False)
except ImportError:
    CPU_PHYSICAL_CORES = (os.cpu_count() or 1) // 2
if not CPU_PHYSICAL_CORES:
    CPU_PHYSICAL_CORES = 1



r"""
QUESTIONS

This extension offers a new syntax to write tests and answers.

An example:

    ~~~~~~~~~~~~~~~~~~~~~
    This is an easy test.
    ~~~~~~~~~~~~~~~~~~~~~

    === QUESTIONS ===

    What's your name ?
    -------------------

    Abraham Lincoln
    ____________________________________________________________________

    How old are you ?

    -------------------

    I'm actually <<<very old>>>.

    ____________________________________________________________________

    Do you like icecream ?

    -------------------

    Yes, pretty much.

    _____________________________________________________________________

    1+1= <<<2>>>

    _____________________________________________________________________

    Complete the multiplication table of binary numbers~:

    ~~~~~~~~~~~~~~~~~~~~~~
    ~~~~~~~~~~~~~~~~~~~~~~
    \begin{tabular}{|c||c|c|}
    \hline
    $\times$ & 0 & 1 \\
    \hline
    \hline
    0        &   &   \\
    \hline
    1        &   &   \\
    \hline
    \end{tabular}
    ~~~~~~~~~~~~~~~~~~~~~~
    ~~~~~~~~~~~~~~~~~~~~~~

    -------------------

    \begin{tabular}{|c||c|c|}
    \hline
    $\times$ & 0 & 1 \\
    \hline
    \hline
    0        & 0 & 0 \\
    \hline
    1        & 0 & 1 \\
    \hline
    \end{tabular}

    ==================


Another simpler example, with no enumeration (only one question):

    ~~~~~~~~~~~~~~~~~~~

    1+1=?

    ---------------

    2

    ~~~~~~~~~~~~~~~~~~~
"""

from re import sub, DOTALL, match
from ptyx.extensions import extended_python


def main(text, compiler):
    text = extended_python.main(text, compiler)

    # <<<text for question (optional):::inline answer>>>
    def inline_answer(m):
        content = m.group("content")
        # Don't use | to split as it is commonly used for absolute values in mathematics.
        question, answer, *remaining = content.split(":::")
        if len(remaining) == 0:
            return f"#QUESTION{{{question}}}#ANSWER{{{answer}}}"
        # It's not clear what to do if there are more than one :::.
        print(f"WARNING: ambiguous syntax {content!r}!")
        return f"#ANSWER{{{content}}}"

    text = sub("<{3,}(?P<content>.*?)>{3,}", inline_answer, text)

    # ~~~~~~~~~~~~~~~~~~
    # ~~~~~~~~~~~~~~~~~~
    # Lonely question or some explications
    # (won't be displayed in the version with answers)
    # ~~~~~~~~~~~~~~~~~~
    # ~~~~~~~~~~~~~~~~~~
    text = sub(
        "\n[ \t]*~{3,}[ \t]*\n[ \t]*~{3,}[ \t]*\n(?P<content>.*?)\n[ \t]*~{3,}[ \t]*\n[ \t]*~{3,}[ \t]*\n",
        "\n#ASK_ONLY\n\\g<content>\n#END\n",
        text,
        flags=DOTALL,
    )

    lines = []
    stack = ["ROOT"]

    def close_any_ANS_ASK_blocks(stack, lines):
        while stack[-1] in ("ASK", "ANS", "ASK_ONLY"):
            print(stack)
            stack.pop()
            lines.append("#END")

    for line in text.split("\n"):
        stripped_line = line.strip()
        n = len(stripped_line)
        if n == 0:
            lines.append(line)  # line, not stripped_line (to keep spaces) !
        elif stripped_line == "-":
            # Blank line for answer.
            lines.append("")
            lines.append(r"\dotfill")
        elif stripped_line == n * "_":
            # New question.
            close_any_ANS_ASK_blocks(stack, lines)
            lines.append("#ITEM")
            lines.append(r"\item")
            lines.append("#ASK")
            stack.append("ASK")
        elif stripped_line == n * "-":
            # New answer.
            close_any_ANS_ASK_blocks(stack, lines)
            lines.append("#ANS")
            stack.append("ANS")
        elif stripped_line == n * ".":
            # A block of python code.
            if stack[-1] == "PYTHON":
                stack.pop()
                lines.append("#END")
            else:
                stack.append("PYTHON")
                lines.append("#PYTHON")
        elif stripped_line == n * "~":
            # A block of text, either a lonely question, an explanation or question+answer.
            # ~~~~~~~~~~~~~~~~~
            # lonely question
            # ---------------
            # corresponding answer
            # ~~~~~~~~~~~~~~~~~
            if stack[-1] in ("ASK", "ANS"):
                lines.append("#END")
                stack.pop()
            else:
                lines.append("#ASK")
                stack.append("ASK")
        elif match("=+[ \t]*((SHUFFLE)|[?]*(!+[?]*)+)[ \t]*=+", stripped_line):
            # Open a group of randomly shuffled questions.
            close_any_ANS_ASK_blocks(stack, lines)
            stack.append("SHUFFLE")
            lines.append(r"\begin{enumerate}")
            lines.append(r"#SHUFFLE")
            lines.append(r"#ITEM")
            lines.append(r"\item")
            lines.append(r"#ASK")
            stack.append("ASK")
        elif match("=+[ \t]*((QUESTIONS)|[?]+)[ \t]*=+", stripped_line):
            # Open a group of questions.
            close_any_ANS_ASK_blocks(stack, lines)
            stack.append("QUESTIONS")
            lines.append(r"\begin{enumerate}")
            lines.append(r"#ENUM")
            lines.append(r"#ITEM")
            lines.append(r"\item")
            lines.append(r"#ASK")
            stack.append("ASK")
        elif stripped_line == n * "=":
            if "QUESTIONS" in stack or "SHUFFLE" in stack:
                # End a group of questions.
                close_any_ANS_ASK_blocks(stack, lines)
                assert stack[-1] in ("QUESTIONS", "SHUFFLE"), stack[-1]
                lines.append("#END")
                lines.append(r"\end{enumerate}")
                stack.pop()
            else:
                # A block of text, either a lonely question, an explanation or question+answer.
                # ===============
                # lonely question
                # ---------------
                # corresponding answer
                # ===============
                if stack[-1] in ("ASK", "ANS"):
                    lines.append("#END")
                    stack.pop()
                else:
                    lines.append("#ASK")
                    stack.append("ASK")
        elif match("[ \t]*[*]+[ \t]*EXERCISE[ \t]*[*]+", stripped_line):
            lines.append(r"\section{}")
            lines.append("#ASK")
            stack.append("ASK")
        else:
            lines.append(line)  # line, not stripped_line (to keep spaces) !
    return "\n".join(lines)



r"""
QUESTIONS

This extension add some useful shortcuts to Python scripts.

An example:


    ...........................
    let a, b
    let a, b in 2..5
    let a, b +
    let c, d, e, f +-
    let a, b /
    let a +
    let b -
    let a +-1
    let c,d in -5..
    let a, b +
    let c, d -
    let u in -3..-1
    let p in 2,3,5,7,11,13,17,19
    let a, b in 2..20 with a > 2*b
    ...........................
    """

from re import sub, DOTALL


def parse_extended_python_code(code):
    """Convert 'extend python' code to pure python code."""
    python_code = code.split("\n")
    for line_number, line in enumerate(python_code):
        if not line.lstrip().startswith("let "):
            continue
        indent = (len(line) - len(line.lstrip())) * " "
        line = line.strip()[4:]
        i = line.find(" with ")
        if i != -1:
            condition = line[i + 6 :]
            line = line[:i]
        else:
            condition = None
        if " in " in line:
            names, val = line.split(" in ")
            # parse val
            if ".." in val:
                a, b = val.split("..")
                if a.startswith("+-") or a.startswith("-+"):
                    a = a[2:]
                    f = "srandint"
                else:
                    f = "randint"
                args = [f, "a=%s" % a.strip(), "b=%s" % b.strip()]
            else:
                f = "randchoice"
                args = [f, "items=[%s]" % val.strip()]

        else:
            if not line:
                raise SyntaxError("lonely `let`.")
            if line.endswith("+-") or line.endswith("-+"):
                # let a, b +-
                args = ["srandint"]
                names = line[:-2]
            elif line.endswith("+"):
                # let a, b +
                args = ["randint"]
                names = line[:-1]
            elif line.endswith("-"):
                # let a, b -
                args = ["randint", "a=-9", "b=-2"]
                names = line[:-1]
            elif line.endswith("/"):
                # let a, b /
                args = ["srandfrac"]
                names = line[:-1]
            else:
                # let a, b
                args = ["srandint"]
                names = line

        names = [name.strip() for name in names.split(",")]
        if not all(name.isidentifier() for name in names):
            raise SyntaxError(f"Line {line!r} not understood.")

        line = "%s, = many(%s, %s)" % (", ".join(names), len(names), ", ".join(args))
        if condition:
            sublines = [f"{indent}while True:"]
            indent += 4 * " "
            sublines.append(indent + line)
            sublines.append(f"{indent}if {condition}:")
            indent += 4 * " "
            sublines.append(f"{indent}break")
            python_code[line_number] = "\n".join(sublines)

        else:
            python_code[line_number] = indent + line
    return "\n".join(python_code)


def main(text, compiler):
    def parse(m):
        content = m.group("content")
        return "\n#PYTHON\n%s\n#END\n" % parse_extended_python_code(content)

    # ............
    # Python code
    # ............
    return sub(
        "\n[ \t]*\\.{4,}[ \t]*\n(?P<content>.*?)\n[ \t]*\\.{4,}[ \t]*\n",
        parse,
        text,
        flags=DOTALL,
    )



from importlib import metadata

# Import ptyx.config first, to set up ptyx !
from ptyx.config import param

__version__ = metadata.version(__package__)
# __version__ = "23.3.0"
__all__ = ("param", "__version__")



"""
Geophyx

Geophar binding for pTyX.

This extension enables to use geophar drawing inside pTyX files.

To load it, use `#LOAD{geophyx}`.

It adds the following commands to pTyX:

    - #CALC{expr} will calculate expression, using a user friendly syntax.
      Example: #CALC{2x²-x(7x-3)}

    - #GEO
      A = Point(2, 5)
      B = Point(0, 1)
      C = Point(2, 7)
      ABC = Triangle(A, B, C)
      #END_GEO

    - #TABVAL
      code (see Geophar doc.)
      #END_TABVAL

    -


"""

# mypy: ignore-errors

import random

from ptyx.extensions import CompilerExtension
from wxgeometrie.modules.tablatex import tabval, tabvar, tabsign
from wxgeometrie.mathlib.parsers import traduire_formule
from wxgeometrie.mathlib.interprete import Interprete
from wxgeometrie.geolib import Feuille

import sympy
from ptyx.latex_generator import LatexGenerator


# ------------------
#    CUSTOM TAGS
# ------------------


class GeophyxLatexGenerator(LatexGenerator):
    def _parse_CALC_tag(self, node):
        args, kw = self._parse_options(node)
        assert len(args) <= 1 and len(kw) == 0
        name = args[0] if args else "RESULT"

        fonctions = [
            key for key, val in self.context.items() if isinstance(val, (type(sympy.sqrt), type(sympy.cos)))
        ]

        def eval_and_store(txt, name):
            formule = traduire_formule(txt, fonctions=fonctions)
            print("Formule interpretation:", txt, " → ", formule)
            self.context[name] = self._eval_python_expr(formule)
            return txt

        self._parse_children(node.children[0].children, function=eval_and_store, name=name)

    def _parse_GCALC_tag(self, node):
        state = random.getstate()
        args, kw = self._parse_options(node)
        for key in kw:
            kw[key] = eval(kw[key])

        def _eval2latex(code):
            print("code::" + repr(code))
            return Interprete(**kw).evaluer(code.strip())[1]

        self._parse_children(node.children, function=_eval2latex, **kw)
        random.setstate(state)

    def _parse_GEO_tag(self, node):
        state = random.getstate()
        args, kw = self._parse_options(node)
        scale = kw.pop("scale", None)
        for key in kw:
            kw[key] = eval(kw[key])

        def _eval2latex(code):
            print("code::" + repr(code))
            feuille = Feuille(**kw)
            for commande in code.split("\n"):
                feuille.executer(commande)
            return feuille.exporter("tikz", echelle=scale)

        self._parse_children(node.children, function=_eval2latex, **kw)
        random.setstate(state)

    def _parse_TABVAL_tag(self, node):
        state = random.getstate()
        args, kw = self._parse_options(node)
        for key in kw:
            kw[key] = eval(kw[key])
        self._parse_children(node.children, function=tabval, **kw)
        random.setstate(state)

    def _parse_TABVAR_tag(self, node):
        state = random.getstate()
        args, kw = self._parse_options(node)
        for key in kw:
            kw[key] = eval(kw[key])
        self._parse_children(node.children, function=tabvar, **kw)
        random.setstate(state)

    def _parse_TABSIGN_tag(self, node):
        state = random.getstate()
        args, kw = self._parse_options(node)
        for key in kw:
            kw[key] = eval(kw[key])
        self._parse_children(node.children, function=tabsign, **kw)
        random.setstate(state)


def extend_compiler() -> CompilerExtension:
    """Function called by the compiler when loading this extension, to add ability to parse new tags."""
    tags = {
        "CALC": (1, 0, None),
        "GCALC": (0, 0, ["@END_GCALC", "@END"]),
        "GEO": (0, 0, ["@END_GEO", "@END"]),
        "TABSIGN": (0, 0, ["@END_TABSIGN", "@END"]),
        "TABVAL": (0, 0, ["@END_TABVAL", "@END"]),
        "TABVAR": (0, 0, ["@END_TABVAR", "@END"]),
    }
    return {"latex_generator": GeophyxLatexGenerator, "tags": tags}


def main(code, compiler):
    # Code is left untouched, this extension only adds new commands.
    return code



from typing import TypedDict, TYPE_CHECKING, Type

if TYPE_CHECKING:
    from ptyx.latex_generator import LatexGenerator


class CompilerExtension(TypedDict):
    latex_generator: Type["LatexGenerator"]
    tags: dict[str, tuple[int, int, None | list[str]]]



"""
EXTENDED PYTHON

"""


from ptyx.extensions.extended_python import main


def test_extended_python():
    text = r"""

................................
let a, b
let c +
let e, f, g, h -
let a in 3..7
let u, v in -5..-2
let p in 2,3,5,7,11,13,17,19
................................
some text
................................
let a,b,c /
................................

"""
    text2 = """

#PYTHON
a, b, = many(2, srandint)
c, = many(1, randint)
e, f, g, h, = many(4, randint, a=-9, b=-2)
a, = many(1, randint, a=3, b=7)
u, v, = many(2, randint, a=-5, b=-2)
p, = many(1, randchoice, items=[2,3,5,7,11,13,17,19])
#END
some text
#PYTHON
a, b, c, = many(3, srandfrac)
#END

"""
    assert main(text, None) == text2


def test_extended_python_with():
    text = r"""
Some text.
................................
def f():
    let a, b with a*b % 2 == 0
    let c,d in 2..5 with c > d
f()
................................
Some text too.
"""
    text2 = """
Some text.
#PYTHON
def f():
    while True:
        a, b, = many(2, srandint)
        if a*b % 2 == 0:
            break
    while True:
        c, d, = many(2, randint, a=2, b=5)
        if c > d:
            break
f()
#END
Some text too.
"""
    assert main(text, None) == text2



#!/usr/bin/env python3

# --------------------------------------
#                  pTyX
#              main script
# --------------------------------------
#    pTyX
#    Python LaTeX preprocessor
#    Copyright (C) 2009-2016  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import argparse
import sys
import csv
from ast import literal_eval
from pathlib import Path

from ptyx import __version__
from ptyx.config import param


if sys.version_info.major == 2:
    raise RuntimeError("Python version 3.10+ is needed !")


class PtyxArgumentParser(argparse.ArgumentParser):
    """Parse pTyX command line interface options."""

    def __init__(self):
        super().__init__(
            prog="pTyX",
            description="Compile .ptyx files into .tex or .pdf files.",
            usage="ptyx [options] filename(s).\nTry 'ptyx --help' for more information.",
        )
        self.add_argument("filenames", nargs="+")
        group = self.add_mutually_exclusive_group()
        group.add_argument(
            "-n",
            "--number-of-documents",
            type=int,
            help="Number of pdf files to generate. Default is %s.\n \
                       Ex: ptyx -n 5 my_file.ptyx"
            % param["total"],
        )
        group.add_argument(
            "--names",
            type=str,
            metavar="CSV_FILE",
            help="Name of a CSV file containing a column of names \
                               (and optionally a second column with forenames). \n \
                               The names will be used to generate the #NAME tag \
                               replacement value.\n \
                               Additionally, if `-n` option is not specified, \
                               default value will be the number of names in the CSV file.",
        )
        self.add_argument(
            "-f",
            "--formats",
            default=param["default_formats"],
            choices=param["formats"],
            help="Output format. Default is %(default)s.\n" "Ex: ptyx -f tex my_file.ptyx",
        )
        self.add_argument(
            "-r",
            "--remove",
            action="store_true",
            help="Remove the `.compile` folder after compilation.",
        )
        self.add_argument("-b", "--debug", action="store_true", help="Debug mode.")
        self.add_argument(
            "-q",
            "--quiet",
            action="store_true",
            help="Suppress most of latex processor output.",
        )
        self.add_argument(
            "-s",
            "--start",
            default=1,
            type=int,
            help="Number of the first generated document \
                       (initial value of internal NUM counter). Default is %(default)s.",
        )
        self.add_argument(
            "-c",
            "--cat",
            action="store_true",
            help="Cat all generated pdf files inside a single one.",
        )
        self.add_argument(
            "-C",
            "--compress",
            action="store_true",
            help="Like --cat, but compress final pdf file using pdf2ps and ps2pdf. "
            "Ghostscript must be present (on Linux, it is probably already installed).",
        )
        self.add_argument(
            "--reorder-pages",
            choices=["brochure", "brochure-reversed"],
            help="Reorder pages for printing.\n\
                Currently, only 'brochure' and 'brochure-reversed' mode are supported.\
                The `pdftk` command must be installed.\n\
                Ex: ptyx --reorder-pages=brochure-reversed -f pdf my_file.ptyx.",
        )
        group2 = self.add_mutually_exclusive_group()
        group2.add_argument(
            "-p",
            "--fixed_number_of_pages",
            metavar="N",
            nargs="?",
            type=int,
            default=argparse.SUPPRESS,
            help="Keep only pdf files whose pages number match N. \
                This may be useful for printing pdf later. \
                Note that the number of documents may not be respected then, so \
                you may have to adjust the number of documents manually.",
        )
        group2.add_argument(
            "-P",
            "--auto-fixed_number_of_pages",
            action="store_true",
            help=(
                "Ensure that all documents have the same number of pages. "
                "The number of documents is respected."
            ),
        )
        self.add_argument(
            "-nc",
            "--no-correction",
            action="store_true",
            help="Don't generate a correction of the test.",
        )
        self.add_argument(
            "-g",
            "--generate-batch-for-windows-printing",
            action="store_true",
            help="Generate a batch file for printing all pdf files using SumatraPDF.",
        )
        self.add_argument(
            "--cpu-cores",
            type=int,
            metavar="N_CORES",
            help="Number of cpu cores to use when compiling. (Automatically detected by default).",
        )
        self.add_argument(
            "--context",
            default="",
            type=str,
            help="Manually customize context (ie. internal namespace).\n \
                       Ex: ptyx --context \"a = 3; b = 2; t = 'hello'\"",
        )
        self.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    def parse_args(self, **kwargs):
        options = super().parse_args(**kwargs)
        options.formats = options.formats.split("+")
        if options.compress or options.cat:
            # pdftk and ghostscript must be installed.
            if "pdf" not in options.formats:
                raise RuntimeError("--cat or --compress option invalid unless pdf output is selected.")
        if options.debug:
            param["debug"] = True
        if options.names:
            with open(Path(options.names).expanduser().resolve()) as f:
                options.names_list = [" ".join(line) for line in csv.reader(f)]
                print("Names extracted from CSV file:")
                print(options.names_list)
            options.number_of_documents = len(options.names)
        else:
            options.names_list = []
        if options.number_of_documents is None:
            options.number_of_documents = param["total"]
        return options


def ptyx(parser=PtyxArgumentParser()):
    """Main pTyX procedure."""

    # First, parse all arguments (filenames, options...)
    # --------------------------------------------------
    options = parser.parse_args()

    if not options.filenames:
        # Exit quickly before main imports, so that `ptyx --help` is fast.
        exit(0)

    from ptyx.compilation import make_files
    from ptyx.latex_generator import compiler

    # TODO: remove kwargs and explicitly pass arguments, to verify types.
    kwargs = vars(options)
    if "fixed_number_of_pages" in kwargs:
        kwargs["pages"] = kwargs["fixed_number_of_pages"]
        kwargs["fixed_number_of_pages"] = True
    else:
        kwargs["fixed_number_of_pages"] = False

    context = {}
    for keyval in kwargs.pop("context", "").split(";"):
        if keyval.strip():
            key, val = keyval.split("=", 1)
            key = key.strip()
            if not str.isidentifier(key):
                raise NameError(f"{key} is not a valid variable name.")
            context[key] = literal_eval(val)

    # Time to act ! Let's compile all ptyx files...
    # ---------------------------------------------

    for input_name in options.filenames:
        # Read pTyX file.
        print(f"Reading {input_name}...")
        input_name = Path(input_name).expanduser().resolve()
        compiler.read_file(input_name)
        # Parse #INCLUDE tags, load extensions if needed, read seed.
        compiler.preparse()

        # Generate syntax tree.
        # The syntax tree is generated only once, and will then be used
        # for all the following compilations.
        compiler.generate_syntax_tree()
        # print(compiler.state['syntax_tree'].display())

        # Compile and generate output files (tex or pdf)
        output_basename, nums = make_files(input_name, **kwargs)

        # Keep track of the seed used.
        seed_value = compiler.seed
        seed_file_name = output_basename.parent / ".seed"
        with open(seed_file_name, "w") as seed_file:
            seed_file.write(str(seed_value))

        # If any of the so-called `ANSWER_tags` is present, compile a second
        # version of the documents with answers.
        # TODO: make an API to choose if the version with answers must be generated or not:
        # - there should be 3 modes, True, False and Auto (current mode).
        # - each mode should be accessible from the command line (add an option)
        # - it should be easy to modify mode for extensions.
        if not options.no_correction:
            ANSWER_tags = ("ANS", "ANSWER", "ASK", "ASK_ONLY")

            tags = compiler.syntax_tree.tags
            if any(tag in tags for tag in ANSWER_tags):
                make_files(input_name, correction=True, _nums=nums, context=context, **kwargs)


if __name__ == "__main__":
    ptyx()
